"""Test suite for basic backend simulation functionalities.
"""

# external imports
from typing import Union

import numpy as np
import pytest

# pyrates internal imports
from pyrates.backend import ComputeGraph
from pyrates.frontend import CircuitTemplate

# meta infos
__author__ = "Richard Gast, Daniel Rose"
__status__ = "Development"


###########
# Utility #
###########


def setup_module():
    print("\n")
    print("==================================================")
    print("| Test Suite: Backend Simulation Functionalities |")
    print("==================================================")


def nmrse(x: np.ndarray,
          y: np.ndarray
          ) -> Union[float, np.ndarray]:
    """Calculates the normalized root mean squared error of two vectors of equal length.
    Parameters
    ----------
    x,y
        Arrays to calculate the nmrse between.
    Returns
    -------
    float
        Normalized root mean squared error.
    """

    max_val = np.max((np.max(x, axis=0), np.max(y, axis=0)))
    min_val = np.min((np.min(x, axis=0), np.min(y, axis=0)))

    diff = x - y

    return np.sqrt(np.sum(diff ** 2, axis=0)) / (max_val - min_val)


#########
# Tests #
#########


def test_2_1_operator():
    """Testing operator functionality of compute graph class:

    See Also
    --------
    :method:`add_operator`: Detailed documentation of method for adding operations to instance of `ComputeGraph`.
    """

    backends = ["tensorflow", "numpy"]
    accuracy = 1e-4

    for b in backends:

        # test correct numerical evaluation of operator with two coupled simple, linear equations
        #########################################################################################

        # create net config from YAML file
        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net0").apply()

        # instantiate compute graph from net config
        dt = 1e-1
        net = ComputeGraph(net_config=net_config, name='net0', vectorization=True, backend=b)

        # simulate operator behavior
        sim_time = 10.0
        results = net.run(simulation_time=sim_time, step_size=dt, outputs={'a': 'pop0/op0/a'})
        net.clear()

        # generate target values
        sim_steps = int(sim_time / dt)
        update0_1 = lambda x: x * 0.5
        update0_0 = lambda x: x + 2.0
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 0] = targets[i, 0] + dt * update0_0(targets[i, 1])
            targets[i + 1, 1] = targets[i, 1] + dt * update0_1(targets[i, 0])

        # compare results with target values
        diff = results['a'].values[:, 0] - targets[1:, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of operator with a single differential equation and external input
        ######################################################################################################

        # set up operator in pyrates
        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net1").apply()
        net = ComputeGraph(net_config=net_config, name='net1', vectorization=True, backend=b)

        # define input
        inp = np.zeros((sim_steps,)) + 0.5

        # simulate operator behavior
        results = net.run(sim_time, step_size=dt, inputs={'pop0/op1/u': inp}, outputs={'a': 'pop0/op1/a'})
        net.clear()

        # calculate operator behavior from hand
        update1 = lambda x, y: x + dt * (y - x)
        targets = np.zeros((sim_steps + 1, 1), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1] = update1(targets[i], inp[i])

        diff = results['a'].values[:, 0] - targets[1:, 0]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of operator with two coupled equations (1 ODE, 1 non-DE eq.)
        ################################################################################################

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net2").apply()
        net = ComputeGraph(net_config=net_config, name='net2', vectorization=True, backend=b)
        results = net.run(sim_time, outputs={'a': 'pop0/op2/a'}, step_size=dt)
        net.clear()

        # calculate operator behavior from hand
        update2 = lambda x: 1. / (1. + np.exp(-x))
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 1] = update2(targets[i, 0])
            targets[i + 1, 0] = update1(targets[i, 0], targets[i + 1, 1])

        diff = results['a'].values[:, 0] - targets[1:, 0]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of operator with a two coupled DEs
        ######################################################################

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net3").apply()
        net = ComputeGraph(net_config=net_config, name='net3', vectorization=True, backend=b)
        results = net.run(sim_time, outputs={'b': 'pop0/op3/b'}, inputs={'pop0/op3/u': inp}, out_dir="/tmp/log",
                          step_size=dt)
        net.clear()

        # calculate operator behavior from hand
        update3_0 = lambda a, b, u: a + dt * (-10. * a + b ** 2 + u)
        update3_1 = lambda b, a: b + dt * 0.1 * a
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 0] = update3_0(targets[i, 0], targets[i, 1], inp[i])
            targets[i + 1, 1] = update3_1(targets[i, 1], targets[i, 0])

        diff = results['b'].values[:, 0] - targets[1:, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)


def test_2_2_node():
    """Testing node functionality of compute graph class.

    See Also
    --------
    :method:`add_node`: Detailed documentation of method for adding nodes to instance of `ComputeGraph`.
    """

    backends = ['numpy', 'tensorflow']

    accuracy = 1e-4

    for b in backends:

        # test correct numerical evaluation of node with 2 operators, where op1 projects to op2
        #######################################################################################

        # set up node in pyrates
        dt = 1e-1
        sim_time = 10.
        sim_steps = int(sim_time / dt)
        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net4").apply()
        net = ComputeGraph(net_config=net_config, name='net0', vectorization=True, backend=b)

        # simulate node behavior
        results = net.run(sim_time, outputs={'a': 'pop0/op1/a'}, step_size=dt)
        net.clear()

        # calculate node behavior from hand
        update0 = lambda x: x + dt * 2.
        update1 = lambda x, y: x + dt * (y - x)
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update1(targets[i, 1], targets[i + 1, 0])

        diff = results['a'].values[:, 0] - targets[:-1, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of node with 2 independent operators
        ########################################################################

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net5").apply()
        net = ComputeGraph(net_config=net_config, name='net1', vectorization=True, backend=b)

        # simulate node behavior
        results = net.run(sim_time, outputs={'a': 'pop0/op5/a'}, step_size=dt)
        net.clear()

        # calculate node behavior from hand
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update1(targets[i, 1], 0.)

        diff = results['a'].values[:, 0] - targets[:-1, 1]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of node with 2 independent operators projecting to the same target operator
        ###############################################################################################################

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net6").apply()
        net = ComputeGraph(net_config=net_config, name='net2', vectorization=True, backend=b)
        results = net.run(sim_time, outputs={'a': 'pop0/op1/a'}, step_size=dt)
        net.clear()

        # calculate node behavior from hand
        targets = np.zeros((sim_steps + 1, 3), dtype=np.float32)
        update2 = lambda x: x + dt * (4. + np.tanh(0.5))
        for i in range(sim_steps):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update2(targets[i, 1])
            targets[i + 1, 2] = update1(targets[i, 2], targets[i + 1, 0] + targets[i + 1, 1])

        diff = results['a'].values[:, 0] - targets[:-1, 2]
        assert np.mean(np.abs(diff)) == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of node with 1 source operator projecting to 2 independent targets
        ######################################################################################################

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net7").apply()
        net = ComputeGraph(net_config=net_config, name='net3', vectorization=True, backend=b)
        results = net.run(sim_time, outputs={'a': 'pop0/op1/a', 'b': 'pop0/op3/b'}, step_size=dt)

        # calculate node behavior from hand
        targets = np.zeros((sim_steps + 1, 4), dtype=np.float32)
        update3 = lambda a, b, u: a + dt * (-10. * a + b ** 2 + u)
        update4 = lambda x, y: x + dt * 0.1 * y
        for i in range(sim_steps):
            targets[i + 1, 0] = update0(targets[i, 0])
            targets[i + 1, 1] = update1(targets[i, 1], targets[i + 1, 0])
            targets[i + 1, 2] = update3(targets[i, 2], targets[i, 3], targets[i + 1, 0])
            targets[i + 1, 3] = update4(targets[i, 3], targets[i, 2])

        diff = np.mean(np.abs(results['a'].values[:, 0] - targets[:-1, 1])) + \
               np.mean(np.abs(results['b'].values[:, 0] - targets[:-1, 3]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)


def test_2_3_edge():
    """Testing edge functionality of compute graph class.

    See Also
    --------
    :method:`add_edge`: Detailed documentation of add_edge method of `ComputeGraph`class.

    """

    backends = ['numpy', 'tensorflow']
    accuracy = 1e-4

    for b in backends:

        # test correct numerical evaluation of graph with 1 source projecting unidirectional to 2 target nodes
        ######################################################################################################

        # set up edge in pyrates
        dt = 1e-1
        sim_time = 10.
        sim_steps = int(sim_time / dt)
        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net8").apply()
        net = ComputeGraph(net_config=net_config, name='net0', vectorization=True, backend=b)

        # calculate edge behavior from hand
        update0 = lambda x, y: x + dt * y * 0.5
        update1 = lambda x, y: x + dt * (y + 2.0)
        update2 = lambda x, y: x + dt * (y - x)
        targets = np.zeros((sim_steps + 1, 4), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 0] = update0(targets[i, 0], targets[i, 1])
            targets[i + 1, 1] = update1(targets[i, 1], targets[i, 0])
            targets[i + 1, 2] = update2(targets[i, 2], targets[i, 0] * 2.0)
            targets[i + 1, 3] = update2(targets[i, 3], targets[i, 0] * 0.5)

        # simulate edge behavior
        results = net.run(sim_time, outputs={'a': 'pop1/op1/a', 'b': 'pop2/op1/a'}, step_size=dt)
        net.clear()

        diff = np.mean(np.abs(results['a'].values[:, 0] - targets[1:, 2])) + \
               np.mean(np.abs(results['b'].values[:, 0] - targets[1:, 3]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of graph with 2 bidirectionaly coupled nodes
        ################################################################################

        # define input
        inp = np.zeros((sim_steps, 1)) + 0.5

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net9").apply()
        net = ComputeGraph(net_config=net_config, name='net1', vectorization=True, backend=b)
        results = net.run(sim_time, outputs={'a': 'pop0/op1/a', 'b': 'pop1/op7/a'}, inputs={'pop1/op7/inp': inp},
                          step_size=dt)
        net.clear()

        # calculate edge behavior from hand
        update3 = lambda x, y, z: x + dt * (y + z - x)
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 0] = update2(targets[i, 0], targets[i, 1] * 0.5)
            targets[i + 1, 1] = update3(targets[i, 1], targets[i, 0] * 2.0, inp[i])

        diff = np.mean(np.abs(results['a'].values[:, 0] - targets[1:, 0])) + \
               np.mean(np.abs(results['b'].values[:, 0] - targets[1:, 1]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of graph with 2 bidirectionally delay-coupled nodes
        #######################################################################################

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net10").apply()
        net = ComputeGraph(net_config=net_config, name='net2', vectorization=True, backend=b, step_size=dt)
        results = net.run(sim_time, outputs={'a': 'pop0/op8/a', 'b': 'pop1/op8/a'}, step_size=dt)
        net.clear()

        # calculate edge behavior from hand
        delay0 = int(0.5 / dt)
        delay1 = int(1. / dt)
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        update4 = lambda x, y: x + dt * (2.0 + y)
        for i in range(sim_steps):
            inp0 = 0. if i < delay0 else targets[i - delay0, 1]
            inp1 = 0. if i < delay1 else targets[i - delay1, 0]
            targets[i + 1, 0] = update4(targets[i, 0], inp0 * 0.5)
            targets[i + 1, 1] = update4(targets[i, 1], inp1 * 2.0)

        diff = np.mean(np.abs(results['a'].values[:, 0] - targets[1:, 0])) + \
               np.mean(np.abs(results['b'].values[:, 0] - targets[1:, 1]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of graph with 2 unidirectionally, multi-delay-coupled nodes
        ###############################################################################################

        # define input
        inp = np.zeros((sim_steps, 1)) + 0.5

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net9").apply()
        net = ComputeGraph(net_config=net_config, name='net3', vectorization=True, step_size=dt, backend=b)
        results = net.run(sim_time, step_size=dt,
                          outputs={'a': 'pop0/op1/a',
                                             'b': 'pop1/op7/a'},
                          inputs={'pop1/op7/inp': inp})
        net.clear()

        # calculate edge behavior from hand
        update3 = lambda x, y, z: x + dt * (y + z - x)
        targets = np.zeros((sim_steps + 1, 2), dtype=np.float32)
        for i in range(sim_steps):
            targets[i + 1, 0] = update2(targets[i, 0], targets[i, 1] * 0.5)
            targets[i + 1, 1] = update3(targets[i, 1], targets[i, 0] * 2.0, inp[i])

        diff = np.mean(np.abs(results['a'].values[:, 0] - targets[1:, 0])) + \
               np.mean(np.abs(results['b'].values[:, 0] - targets[1:, 1]))
        assert diff == pytest.approx(0., rel=accuracy, abs=accuracy)

        # test correct numerical evaluation of graph with delay distributions
        #####################################################################

        # define input
        dt = 1e-2
        sim_time = 100.
        sim_steps = int(np.round(sim_time / dt, decimals=0))
        inp = np.zeros((sim_steps, 1)) + 0.5

        net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net13").apply(label='net4')
        net = net_config.compile(vectorization=True, step_size=dt, backend=b, solver='euler')
        results = net.run(sim_time,
                          outputs={'a1': 'p1/op9/a',
                                   'a2': 'p2/op9/a'},
                          inputs={'p1/op9/I_ext': inp})
        # TODO: add evaluation of network behavior by hand and compare results against that
        net.clear()


@pytest.mark.skip
def test_2_4_vectorization():
    """Testing vectorization functionality of ComputeGraph class.

    See Also
    --------
    :method:`_vectorize`: Detailed documentation of vectorize method of `ComputeGraph` class.
    """

    backends = ['tensorflow', 'numpy']
    for b in backends:
        # test whether vectorized networks produce same output as non-vectorized backend
        ################################################################################

        # define simulation params
        dt = 1e-2
        sim_time = 10.
        sim_steps = int(sim_time / dt)
        inp = np.zeros((sim_steps, 2)) + 0.5

        # set up networks
        net_config0 = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net12").apply()
        net_config1 = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net12").apply()
        net0 = ComputeGraph(net_config=net_config0, name='net0', vectorization='none', dt=dt, build_in_place=False,
                            backend=b)
        net1 = ComputeGraph(net_config=net_config1, name='net1', vectorization='nodes', dt=dt, build_in_place=False,
                            backend=b)

        # simulate network behaviors
        results0 = net0.run(sim_time, outputs={'a': 'pop0/op7/a', 'b': 'pop1/op7/a'},
                            inputs={'all/op7/inp': inp})
        results1 = net1.run(sim_time, outputs={'a': 'pop0/op7/a', 'b': 'pop1/op7/a'},
                            inputs={'all/op7/inp': inp}, out_dir='/tmp/log')

        error1 = nmrse(results0.values, results1.values)

        assert np.sum(results1.values) > 0.
        assert np.mean(error1) == pytest.approx(0., rel=1e-6, abs=1e-6)


def test_2_5_solver():
    """Testing different numerical solvers of pyrates.

    See Also
    --------
    :method:`_solve`: Detailed documentation of how to numerical integration is performed by the `NumpyBackend`.
    :method:`run`: Detailed documentation of the method that needs to be called to solve differential equations in the
    `NumpyBackend`.
    """

    backend = 'numpy'

    # define input
    dt = 1e-3
    sim_time = 100.
    sim_steps = int(np.round(sim_time / dt, decimals=0))
    inp = np.zeros((sim_steps, 1)) + 0.5

    # standard euler solver (trusted)
    net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net13").apply(label='net0')
    net = net_config.compile(vectorization=True, step_size=dt, backend=backend, solver='euler')
    results = net.run(sim_time,
                      outputs={'a1': 'p1/op9/a',
                               'a2': 'p2/op9/a'},
                      inputs={'p1/op9/I_ext': inp})
    net.clear()

    # scipy solver (tested)
    net_config2 = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net13").apply(label='net1')
    net2 = net_config2.compile(vectorization=True, step_size=dt, backend=backend, solver='scipy')
    results2 = net2.run(sim_time,
                        outputs={'a1': 'p1/op9/a',
                                 'a2': 'p2/op9/a'},
                        inputs={'p1/op9/I_ext': inp},
                        method='RK23')
    net2.clear()

    assert np.mean(results.loc[:, 'a2'].values - results2.loc[:, 'a2'].values) == pytest.approx(0., rel=1e-4, abs=1e-4)


def test_2_6_inputs_outputs():
    """Tests the input-output interface of the run method in circuits of different hierarchical depth.

    See Also
    -------
    :method:`CircuitIR.run` detailed documentation of how to use the arguments `inputs` and `outputs`.

    """

    backend = 'numpy'
    dt = 1e-3
    sim_time = 100.
    sim_steps = int(np.round(sim_time / dt, decimals=0))

    # define inputs and outputs for each population separately
    ##########################################################

    # define input
    inp1 = np.zeros((sim_steps, 1)) + 0.5
    inp2 = np.zeros((sim_steps, 1)) + 0.2

    # perform simulation
    net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net13").apply(label='net1')
    net = net_config.compile(vectorization=True, step_size=dt, backend=backend, solver='scipy')
    r1 = net.run(sim_time,
                 outputs={'a1': 'p1/op9/a', 'a2': 'p2/op9/a'},
                 inputs={'p1/op9/I_ext': inp1, 'p2/op9/I_ext': inp2})
    net.clear()

    # define input and output for both populations simultaneously
    #############################################################

    backend = 'numpy'

    # define input
    inp = np.zeros((sim_steps, 2))
    inp[:, 0] = 0.5
    inp[:, 1] = 0.2

    # perform simulation
    net_config = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net13").apply(label='net2')
    net = net_config.compile(vectorization=True, step_size=dt, backend=backend, solver='scipy')
    r2 = net.run(sim_time, outputs={'a': 'all/op9/a'}, inputs={'all/op9/I_ext': inp})
    net.clear()

    assert np.mean(r1.values.flatten() - r2.values.flatten()) == pytest.approx(0., rel=1e-4, abs=1e-4)

    # repeat in a network with 2 hierarchical levels of node organization
    #####################################################################

    # define input
    inp3 = np.zeros((sim_steps, 1)) + 0.1
    inp4 = np.zeros((sim_steps, 1))

    # perform simulation
    nc1 = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net14").apply(label='net3')
    n1 = nc1.compile(vectorization=True, step_size=dt, backend=backend, solver='scipy')
    r1 = n1.run(sim_time,
                outputs={'a1': 'c1/p1/op9/a', 'a2': 'c1/p2/op9/a', 'a3': 'c2/p1/op9/a', 'a4': 'c2/p2/op9/a'},
                inputs={'c1/p1/op9/I_ext': inp1, 'c1/p2/op9/I_ext': inp2, 'c2/p1/op9/I_ext': inp3,
                        'c2/p2/op9/I_ext': inp4})
    n1.clear()

    # define input
    inp = np.zeros((sim_steps, 4))
    inp[:, 0] = 0.5
    inp[:, 1] = 0.2
    inp[:, 2] = 0.1

    # perform simulation
    nc2 = CircuitTemplate.from_yaml("model_templates.test_resources.test_backend.net14").apply(label='net4')
    n2 = nc2.compile(vectorization=True, step_size=dt, backend=backend, solver='scipy')
    r2 = n2.run(sim_time, outputs={'a': 'all/all/op9/a'}, inputs={'all/all/op9/I_ext': inp})
    n2.clear()

    assert np.mean(r1.values.flatten() - r2.values.flatten()) == pytest.approx(0., rel=1e-4, abs=1e-4)



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""Contains wrapper classes for different backends that are needed by the parser module.

A new backend needs to implement the following methods:
- __init__
- run
- add_var
- add_op
- add_layer

Currently supported backends:
- Numpy: NumpyBackend.
- Tensorflow: TensorflowBackend.
- Fortran: FortranBackend (experimental).

"""

# external imports
import time
from typing import Optional, Dict, List, Union, Any, Callable
import numpy as np
from copy import deepcopy
import os
import sys
from shutil import rmtree
import warnings
from scipy.interpolate.interpolate import interp1d

# pyrates internal imports
from .funcs import *
from .parser import replace


class NumpyVar(np.ndarray):
    """Base class for adding variables to the PyRates compute graph. Creates a numpy array with additional attributes
    for variable identification/retrieval from graph. Should be used as parent class for custom variable classes.

    Parameters
    ----------
    vtype
        Type of the variable. Can be either `constant` or `state_variable`. Constant variables are necessary to perform
        certain graph optimizations previous to run time.
    dtype
        Data-type of the variable. For valid data-types, check the documentation of the backend in use.
    shape
        Shape of the variable.
    value
        Value of the variable. If scalar, please provide the shape in addition.
    name
        Full name of the variable (including the node and operator it belongs to).
    short_name
        Name of the variable excluding its node and operator.

    Returns
    -------
    NumpyVar
        Instance of NumpyVar.
    """

    def __new__(cls, vtype: str, backend: Any, name: str, dtype: Optional[str] = None,
                shape: Optional[tuple] = None, value: Optional[Any] = None, squeeze: bool = True):
        """Creates new instance of NumpyVar.
        """

        # check whether necessary arguments were provided
        if all([arg is None for arg in [shape, value, dtype]]):
            raise ValueError('Either `value` or `shape` and `dtype` need to be provided')

        # get shape
        if not shape:
            shape = value.shape if hasattr(value, 'shape') else np.shape(value)

        # get data type
        if not dtype:
            if hasattr(value, 'dtype'):
                dtype = value.dtype
            else:
                try:
                    dtype = np.dtype(value)
                except TypeError:
                    dtype = type(value)
        dtype = dtype.name if hasattr(dtype, 'name') else str(dtype)
        dtype = dtype.name if hasattr(dtype, 'name') else str(dtype)
        if dtype in backend.dtypes:
            dtype = backend.dtypes[dtype]
        else:
            for dtype_tmp in backend.dtypes:
                if dtype_tmp in dtype:
                    dtype = backend.dtypes[dtype_tmp]
                    break
            else:
                dtype = backend._float_def
                warnings.warn(f'WARNING! Unknown data type of variable {name}: {dtype}. '
                              f'Datatype will be set to default type: {dtype}.')

        # create variable
        if vtype == 'state_var' and 1 in tuple(shape) and squeeze:
            idx = tuple(shape).index(1)
            shape = list(shape)
            shape.pop(idx)
            shape = tuple(shape)
        if callable(value):
            obj = value
        else:
            value = cls._get_value(value, dtype, shape)
            if squeeze:
                value = cls.squeeze(value)
            obj = cls._get_var(value, name, dtype)

        # store additional attributes on variable object
        obj.short_name = name.split('/')[-1]
        if not hasattr(obj, 'name'):
            obj.name = name
        else:
            name = obj.name
        obj.vtype = vtype

        return obj, name

    def numpy(self):
        """Returns current value of NumpyVar.
        """
        try:
            return self[:]
        except IndexError:
            return self

    @staticmethod
    def _get_value(value, dtype, shape):
        """Defines initial value of variable.
        """
        if value is None:
            return np.zeros(shape=shape, dtype=dtype)
        elif not hasattr(value, 'shape'):
            if type(value) is list:
                return np.zeros(shape=shape, dtype=dtype) + np.asarray(value, dtype=dtype).reshape(shape)
            else:
                return np.zeros(shape=shape, dtype=dtype) + value
        elif shape:
            if value.shape == shape:
                return value
            elif sum(shape) < sum(value.shape):
                return value.squeeze()
            else:
                idx = ",".join("None" if s == 1 else ":" for s in shape)
                return eval(f'value[{idx}]')
        else:
            return np.asarray(value, dtype=dtype)

    @staticmethod
    def squeeze(var, axis=None):
        return var.squeeze(axis)

    @classmethod
    def _get_var(cls, value, name, dtype):
        """Creates new numpy array from NumpyVar.
        """
        return np.array(value).view(cls)

    def __deepcopy__(self, memodict={}):
        obj = super().__deepcopy__(memodict)
        if not hasattr(obj, 'name'):
            obj.name = self.name
        if not hasattr(obj, 'short_name'):
            obj.short_name = self.short_name
        if not hasattr(obj, 'vtype'):
            obj.vtype = self.vtype
        return obj

    def __str__(self):
        return self.name

    def __hash__(self):
        return hash(str(self))

    @staticmethod
    def __subclasscheck__(subclass):
        if np.ndarray.__subclasscheck__(subclass):
            return True
        else:
            return interp1d.__subclasscheck__(subclass)


class PyRatesOp:
    """Base class for adding operations on variables to the PyRates compute graph. Should be used as parent class for
    custom operator classes.

    Parameters
    ----------
    op
        Call-signature of the operation/function.
    short_name
        Name of the operation within the equations.
    name
        Name of the operation/function.
    args
        Arguments to the function.

    """

    var_class = NumpyVar

    def __init__(self, op: str, short_name: str, name, *args, **kwargs) -> None:
        """Instantiates PyRates operator.
        """

        # set identifiers of operator
        self.op = op
        self.short_name = short_name if short_name in name else name.split('/')[-1]
        self.name = name

        # generate function string
        self._op_dict = self.generate_op_str(op, args, **kwargs)

        # extract information from parsing results
        self.value = self._op_dict['value']
        self.arg_names, args = self._get_unique_args(self._op_dict['arg_names'], self._op_dict['args'])
        self.input_ops = self._op_dict['input_ops']
        self.is_constant = self._op_dict['is_constant']

        # build final op
        self.shape = ()
        self.dtype = 'float'
        self._callable = lambda x: x
        self.args = ()
        self.build_op(args)

    def numpy(self):
        """Evaluates the return values of the PyRates operation.
        """
        result = self._callable(*self.args)
        self._check_numerics(result, self.name)
        return result

    def _generate_func(self):
        """Generates a function from operator value and arguments"""
        func_dict = {}
        func = CodeGen()
        func.add_code_line(f"def {self.short_name}(")
        for arg in self._op_dict['arg_names']:
            func.add_code_line(f"{arg},")
        if len(self._op_dict['arg_names']) > 0:
            func.code[-1] = func.code[-1][:-1]
        func.add_code_line("):")
        func.add_linebreak()
        func.add_indent()
        func.add_code_line(f"return {self._op_dict['value']}")
        exec(func.generate(), globals(), func_dict)
        return func_dict

    def build_op(self, args):
        """Builds the function of the operator, tests its functionality and stores shape and data-type of the result
        on the instance.
        """

        func_dict = self._generate_func()
        self._callable = func_dict.pop(self.short_name)

        # test function
        self.args = self._deepcopy(args)
        result = self.numpy() if 'no_op' not in self.short_name else self.args[0]
        self.args = args

        # remember output shape and data-type
        self.shape = result.shape if hasattr(result, 'shape') else ()
        self.dtype = result.dtype if hasattr(result, 'dtype') else type(result)

    @classmethod
    def generate_op_str(cls, op, args, **kwargs):
        """Generates the function string, call signature etc.
        """

        # collect operator arguments
        results = {'value': None, 'args': [], 'arg_names': [], 'is_constant': False, 'input_ops': []}

        if op == 'no_op':
            if hasattr(args[0], 'short_name'):
                val = args[0].short_name
                arg = [args[0]]
                name = [args[0].short_name]
            else:
                val = str(args[0])
                arg = []
                name = []
            results['value'] = val
            results['args'] = arg
            results['arg_names'] = name
            return results

        results, results_args, results_arg_names, n_vars = cls._process_args(args, results)

        # setup operator call
        #####################

        eval_gen = CodeGen()

        # check whether operator is a simple mathematical symbol or involves a function call
        if op in "+-**/*^%<>==>=<=" and len(args) == 2:
            arg1 = cls._arg_to_str(op, args[0], results_args[0], results_arg_names[0])
            arg2 = cls._arg_to_str(op, args[1], results_args[1], results_arg_names[1])
            eval_gen.add_code_line(f"{arg1} {op} {arg2}")
            func_call = False
        else:
            func_call = True
            eval_gen.add_code_line(f"{op}(")

        # add operator arguments
        for i, (key, arg) in enumerate(zip(results_arg_names, results_args)):
            if key != '__no_name__':
                if type(arg) is str:
                    if func_call:
                        if arg == "(":
                            eval_gen.add_code_line(f"{arg}")
                        else:
                            eval_gen.add_code_line(f"{arg},")
                    idx = cls._index(results['args'], arg)
                    results['args'].pop(idx)
                    results['arg_names'].pop(idx)
                elif type(arg) is dict and func_call:
                    eval_gen.add_code_line(f"{arg['value']},")
                elif func_call:
                    eval_gen.add_code_line(f"{key},")
            else:
                if func_call:
                    eval_gen.add_code_line(f"{arg},")
                if key in results['arg_names']:
                    idx = cls._index(results['arg_names'], key)
                    results['args'].pop(idx)
                    results['arg_names'].pop(idx)

        # add function end
        if func_call:
            eval_gen.code[-1] = eval_gen.code[-1][:-1]
            eval_gen.add_code_line(")")
        results['value'] = eval_gen.generate()

        # check whether operation arguments contain merely constants
        if n_vars == 0:
            results['is_constant'] = True

        return results

    @classmethod
    def _process_args(cls, args, results, constants_to_num=True):
        """Parses arguments to function into argument names that are added to the function call and argument values
        that can used to set up the return line of the function.
        """

        results_args = []
        results_arg_names = []
        n_vars = 0

        for idx, arg in enumerate(args):

            if PyRatesOp.__subclasscheck__(type(arg)):

                # extract arguments from other operations and nest them into this operation
                results['input_ops'].append(arg.name)
                pop_indices = []
                for arg_tmp in arg.arg_names:
                    if arg_tmp in results['arg_names']:
                        pop_indices.append(arg.arg_names.index(arg_tmp))
                new_args = arg.args.copy()
                new_arg_names = arg.arg_names.copy()
                for i, pop_idx in enumerate(pop_indices):
                    new_args.pop(pop_idx - i)
                    new_arg_names.pop(pop_idx - i)
                results_args.append({'args': new_args, 'value': arg.lhs if 'assign' in arg.short_name else arg.value,
                                     'arg_names': new_arg_names})
                results_arg_names.append(arg.short_name)
                results['args'] += new_args
                results['arg_names'] += new_arg_names
                n_vars += 1

            elif cls.var_class.__subclasscheck__(type(arg)):

                # parse PyRates variable into the function
                if arg.vtype is 'constant' and not tuple(arg.shape) and constants_to_num:
                    arg_name = '__no_name__'
                    arg_value = arg.numpy()  #if hasattr(arg, 'numpy') else arg
                else:
                    arg_name = arg.short_name
                    arg_value = arg
                    n_vars += 1
                results_args.append(arg_value)
                results_arg_names.append(arg_name)
                results['args'].append(arg_value)
                results['arg_names'].append(arg_name)

            elif type(arg) is tuple or type(arg) is list:

                # parse list of arguments
                tuple_begin = '('
                results_args.append(tuple_begin)
                results_arg_names.append(tuple_begin)
                results['args'].append(tuple_begin)
                results['arg_names'].append(tuple_begin)
                results, results_args_tmp, results_arg_names_tmp, n_vars_tmp = cls._process_args(arg, results)
                results_args += results_args_tmp
                results_arg_names += results_arg_names_tmp
                n_vars += n_vars_tmp
                tuple_end = ')'
                results_args.append(tuple_end)
                results_arg_names.append(tuple_end)
                results['args'].append(tuple_end)
                results['arg_names'].append(tuple_end)

            else:

                # parse strings and constant arguments into the function
                arg_name = "__no_name__"
                results_args.append(arg)
                results_arg_names.append(arg_name)
                results['args'].append(arg)
                results['arg_names'].append(arg_name)

        return results, results_args, results_arg_names, n_vars

    @staticmethod
    def _get_unique_args(arg_names, args):
        arg_names_unique = np.unique(arg_names)
        for name in arg_names_unique:
            n = arg_names.count(name)
            while n > 1:
                i0 = arg_names.index(name)+1
                i1 = arg_names[i0:].index(name)
                arg_names.pop(i0+i1)
                args.pop(i0+i1)
                n -= 1
        return arg_names, args

    @staticmethod
    def _check_numerics(vals: NumpyVar, name: str):
        """Checks whether function evaluation leads to any NaNs or infinite values.
        """
        check = []
        try:
            if hasattr(vals, 'shape'):
                vals = vals.numpy().flatten() if hasattr(vals, 'numpy') else vals.flatten()
            else:
                vals = [vals]
            for val in vals:
                check.append(np.isnan(val) or np.isneginf(val))
        except TypeError:
            check.append(np.isnan(vals) or np.isneginf(vals))
        if any(check):
            raise ValueError(f'Result of operation ({name}) contains NaNs or infinite values.')

    @staticmethod
    def _get_arg(arg: Any, arg_reduced: Any, key: str) -> str:
        if key == '__no_name__':
            return f"{arg_reduced}" if arg_reduced > 0 else f"({arg_reduced})"
        elif PyRatesOp.__subclasscheck__(type(arg)):
            return f"{arg.value}"
        else:
            return key

    @classmethod
    def _arg_to_str(cls, op: str, arg, arg_reduced, key: str) -> str:
        op_orders = [['+', '-'],
                     ['*'],
                     ['/', '%', '>', '<', '==', '<=', '>=', '!='],
                     ['^', '**']]
        op_order = 0
        for i, ops in enumerate(op_orders):
            if op in ops:
                op_order = i+1
                break
        arg_order = 5
        if hasattr(arg, 'op'):
            for i, ops in enumerate(op_orders):
                if arg.op in ops:
                    arg_order = i+1
                    break
        arg_str = cls._get_arg(arg, arg_reduced, key)
        if (op_order > arg_order or (op_order == arg_order and op_order >= 3)) and \
                not (arg_str[0] == "(" and arg_str[-1] == ")"):
            return f"({arg_str})"
        return arg_str

    @staticmethod
    def _deepcopy(x):
        return deepcopy(x)

    @staticmethod
    def _index(x, y):
        return x.index(y)


class PyRatesAssignOp(PyRatesOp):
    """Sub-class for assign operations. Typically, an update (2. entry in args) is assigned to a PyRates variable
    (1. entry in args). An index can be provided as third argument to target certain entries of the variable.
    """

    def __init__(self, op: str, short_name: str, name, *args, **kwargs) -> None:
        super().__init__(op, short_name, name, *args, **kwargs)
        self.lhs = self._op_dict.pop('lhs')
        self.rhs = self._op_dict.pop('rhs')

    def _generate_func(self):
        func_dict = {}
        func = CodeGen()
        func.add_code_line(f"def {self.short_name}(")
        for arg in self._op_dict['arg_names']:
            func.add_code_line(f"{arg},")
        func.code[-1] = func.code[-1][:-1]
        func.add_code_line("):")
        func.add_linebreak()
        func.add_indent()
        func.add_code_line(f"{self._op_dict['value']}")
        func.add_linebreak()
        func.add_code_line(f"return {self._op_dict['lhs']}")
        exec(func.generate(), globals(), func_dict)
        return func_dict

    @classmethod
    def generate_op_str(cls, op, args, **kwargs):
        """Generates the assign function string, call signature etc.
        """

        # initialization
        results = {'value': None, 'args': [], 'arg_names': [], 'is_constant': False, 'input_ops': [], 'lhs': None}
        results_args = []
        results_arg_names = []

        # extract variables
        ###################

        var, upd = args[0:2]

        if len(args) > 2:
            var_idx, results_args, results_arg_names = cls._extract_var_idx(op, args, results_args, results_arg_names,
                                                                            **kwargs)
        else:
            var_idx = ""

        # add variable to the function arguments
        var_key = var.short_name if 'assign' not in var.short_name else '__no_name__'
        var_pos = len(results_args)
        results_args.append(var if 'assign' not in var.short_name else {'value': var.lhs})
        results_arg_names.append(var_key)

        # add variable update to the function arguments
        if hasattr(upd, 'vtype') and upd.vtype == 'constant' and not upd.shape:
            upd_key = '__no_name__'
            if not PyRatesOp.__subclasscheck__(type(upd)):
                upd = {'value': upd}
        else:
            upd_key = upd.short_name
        upd_pos = len(results_args)
        results_args.append(upd)
        results_arg_names.append(upd_key)

        # setup function head
        #####################

        # add arguments
        for idx, (key, arg) in enumerate(zip(results_arg_names, results_args)):
            if PyRatesOp.__subclasscheck__(type(arg)):
                pop_indices = []
                for arg_tmp in arg.arg_names:
                    if arg_tmp in results['arg_names']:
                        pop_indices.append(arg.arg_names.index(arg_tmp))
                new_args = arg.args.copy()
                new_arg_names = arg.arg_names.copy()
                for i, pop_idx in enumerate(pop_indices):
                    new_args.pop(pop_idx - i)
                    new_arg_names.pop(pop_idx - i)
                results_args[idx] = {'args': new_args, 'value': arg.value, 'arg_names': new_arg_names}
                results['input_ops'].append(arg.name)
                results['args'] += new_args
                results['arg_names'] += new_arg_names
            elif key != '__no_name__':
                results['args'].append(arg)
                results['arg_names'].append(key)

        # assign update to variable and return variable
        ###############################################

        var_str = results_args[var_pos]['value'] if type(results_args[var_pos]) is dict else var_key
        upd_str = results_args[upd_pos]['value'] if type(results_args[upd_pos]) is dict else upd_key
        if op in ["=", "+=", "-=", "*=", "/="]:
            assign_line = f"{var_str}{var_idx} {op} {upd_str}"
        elif "scatter" in op:
            scatter_into_first_dim = args[-1]
            if len(args) > 3 and scatter_into_first_dim:
                assign_line = f"{var_str}.{op}([{var_idx}], [{upd_str}])"
            else:
                assign_line = f"{var_str}.{op}({var_idx}{upd_str})"
        else:
            assign_line = f"{var_str}{var_idx}.{op}({upd_str})"
        results['value'] = assign_line
        results['lhs'] = var_str
        results['rhs'] = upd_str

        return results

    @classmethod
    def _extract_var_idx(cls, op, args, results_args, results_arg_names, idx_l='[', idx_r=']'):

        if hasattr(args[2], 'vtype'):

            # for indexing via PyRates variables
            key = args[2].short_name
            if hasattr(args[2], 'value'):
                var_idx = f"{idx_l}{args[2].value}{idx_r}"
            else:
                var_idx = f"{idx_l}{key}{idx_r}"

        elif type(args[2]) is str and len(args) > 3:

            # for indexing via string-based indices
            key = args[3].short_name
            var_idx = f"{idx_l}{args[2]}{idx_r}"
            var_idx = var_idx.replace(args[3].short_name, key)

        else:

            # for indexing via integers
            var_idx = f"{idx_l}{args[2]}{idx_r}"
            key = "__no_name__"

        if type(args[2]) is str and len(args) > 3:
            results_args.append(args[3])
        else:
            results_args.append(args[2])
        results_arg_names.append(key)

        return var_idx, results_args, results_arg_names


class PyRatesIndexOp(PyRatesOp):
    """Sub-class for indexing operations. Typically, an index is provided (2. entry in args) indicating which values to
     extract from a variable (1. entry in args).
    """

    @classmethod
    def generate_op_str(cls, op, args, idx_l='[', idx_r=']'):
        """Generates the index function string, call signature etc.
        """

        # initialization
        results = {'value': None, 'args': [], 'arg_names': [], 'is_constant': False, 'shape': (), 'dtype': 'float32',
                   'input_ops': []}

        # extract variable
        ##################

        var_tmp = args[0]
        n_vars = 0

        if PyRatesOp.__subclasscheck__(type(var_tmp)):

            # nest pyrates operations and their arguments into the indexing operation
            var = var_tmp.lhs if hasattr(var_tmp, 'lhs') else var_tmp.value
            pop_indices = []
            for arg_tmp in var_tmp.arg_names:
                if arg_tmp in results['arg_names']:
                    pop_indices.append(var_tmp.arg_names.index(arg_tmp))
            new_args = var_tmp.args.copy()
            new_arg_names = var_tmp.arg_names.copy()
            for i, pop_idx in enumerate(pop_indices):
                new_args.pop(pop_idx - i)
                new_arg_names.pop(pop_idx - i)
            results['input_ops'].append(var_tmp.name)
            results['args'] += new_args
            results['arg_names'] += new_arg_names
            n_vars += 1

        elif cls.var_class.__subclasscheck__(type(var_tmp)):

            # parse pyrates variables into the indexing operation
            if var_tmp.vtype != 'constant' or tuple(var_tmp.shape):
                var = var_tmp.short_name
                results['args'].append(var_tmp)
                results['arg_names'].append(var)
                n_vars += 1
            else:
                var = var_tmp.numpy()

        else:

            raise ValueError(f'Variable type for indexing not understood: {var_tmp}. '
                             f'Please consider another form of variable declaration or indexing.')

        # extract idx
        #############

        idx = args[1]

        if PyRatesOp.__subclasscheck__(type(idx)):

            # nest pyrates operations and their arguments into the indexing operation
            var_idx = f"{idx_l}{idx.lhs if hasattr(idx, 'lhs') else idx.value}{idx_r}"
            pop_indices = []
            for arg_tmp in idx.arg_names:
                if arg_tmp in results['arg_names']:
                    pop_indices.append(idx.arg_names.index(arg_tmp))
            new_args = idx.args.copy()
            new_arg_names = idx.arg_names.copy()
            for i, pop_idx in enumerate(pop_indices):
                new_args.pop(pop_idx - i)
                new_arg_names.pop(pop_idx - i)
            results['input_ops'].append(idx.name)
            results['args'] += new_args
            results['arg_names'] += new_arg_names
            n_vars += 1

        elif cls.var_class.__subclasscheck__(type(idx)):

            # parse pyrates variables into the indexing operation
            key = idx.short_name
            if idx.vtype != 'constant' or tuple(idx.shape):
                var_idx = f"{idx_l}{key}{idx_r}"
                results['args'].append(idx)
                results['arg_names'].append(key)
                n_vars += 1
            else:
                var_idx = f"{idx_l}{idx.numpy()}{idx_r}"

        elif type(idx) is str or "int" in str(type(idx)) or "float" in str(type(idx)):

            # parse constant numpy-like indices into the indexing operation
            pop_indices = []
            for i, arg in enumerate(args[2:]):
                if hasattr(arg, 'short_name') and arg.short_name in idx:
                    if hasattr(arg, 'vtype') and arg.vtype == 'constant' and sum(arg.shape) < 2:
                        idx = replace(idx, arg.short_name, f"{int(arg)}")
                        pop_indices.append(i+2)
                    elif hasattr(arg, 'value'):
                        idx = replace(idx, arg.short_name, arg.value)
                        pop_indices.append(i+2)
                        results['args'] += arg.args
                        results['arg_names'] += arg.arg_names
            var_idx = f"{idx_l}{idx}{idx_r}"
            args = list(args)
            for idx in pop_indices[::-1]:
                args.pop(idx)
            args = tuple(args)

        elif type(idx) in (list, tuple):

            # parse list of constant indices into the operation
            var_idx = f"{list(idx)}"

        else:
            raise ValueError(f'Index type not understood: {idx}. Please consider another form of variable indexing.')

        # setup function head
        #####################

        # remaining arguments
        results, results_args, results_arg_names, n_vars_tmp = cls._process_args(args[2:], results)
        n_vars += n_vars_tmp

        # apply index to variable and return variable
        #############################################

        # generate op
        results['value'] = f"{var}{var_idx}"

        # check whether any state variables are part of the indexing operation or not
        if n_vars == 0:
            results['is_constant'] = True

        return results


class NumpyBackend(object):
    """Wrapper to numpy. This class provides an interface to all numpy functionalities that may be accessed via pyrates.
    All numpy variables and operations will be stored in a layered compute graph that can be executed to evaluate the
    (dynamic) behavior of the graph.

    Parameters
    ----------
    ops
        Additional operations this backend instance can perform, defined as key-value pairs. The key can then be used in
        every equation parsed into this backend. The value is a dictionary again, with two keys:
            1) name - the name of the function/operation (used during code generation)
            2) call - the call signature of the function including import abbreviations (i.e. `np.add` for numpy's add
            function)
    dtypes
        Additional data-types this backend instance can use, defined as key-value pairs.
    name
        Name of the backend instance. Used during code generation to create a recognizable file structure.
    float_default_type
        Default float precision. If no data-type is indicated for a particular variable, this will be used.
    imports
        Can be used to pass additional import statements that are needed for code generation of the custom functions
        provided via `ops`. Will be added to the top of each generated code file.
    build_dir
        Directory in which pyrates builds will be stored.

    """

    idx_l, idx_r = "[", "]"
    idx_start = 0

    def __init__(self,
                 ops: Optional[Dict[str, str]] = None,
                 dtypes: Optional[Dict[str, object]] = None,
                 name: str = 'net_0',
                 float_default_type: str = 'float32',
                 imports: Optional[List[str]] = None,
                 build_dir: str = None,
                 ) -> None:
        """Instantiates numpy backend, i.e. a compute graph with numpy operations.
        """

        super().__init__()

        # define operations and datatypes of the backend
        ################################################

        # base math operations
        self.ops = {"+": {'name': "numpy_add", 'call': "np.add"},
                    "-": {'name': "numpy_subtract", 'call': "np.subtract"},
                    "*": {'name': "numpy_multiply", 'call': "np.multiply"},
                    "/": {'name': "numpy_divide", 'call': "np.divide"},
                    "%": {'name': "numpy_modulo", 'call': "np.mod"},
                    "^": {'name': "numpy_power", 'call': "np.power"},
                    "**": {'name': "numpy_power_float", 'call': "np.float_power"},
                    "@": {'name': "numpy_dot", 'call': "np.dot"},
                    ".T": {'name': "numpy_transpose", 'call': "np.transpose"},
                    ".I": {'name': "numpy_invert", 'call': "np.invert"},
                    ">": {'name': "numpy_greater", 'call': "np.greater"},
                    "<": {'name': "numpy_less", 'call': "np.less"},
                    "==": {'name': "numpy_equal", 'call': "np.equal"},
                    "!=": {'name': "numpy_not_equal", 'call': "np.not_equal"},
                    ">=": {'name': "numpy_greater_equal", 'call': "np.greater_equal"},
                    "<=": {'name': "numpy_less_equal", 'call': "np.less_equal"},
                    "=": {'name': "assign", 'call': "="},
                    "+=": {'name': "assign_add", 'call': "+="},
                    "-=": {'name': "assign_subtract", 'call': "-="},
                    "*=": {'name': "assign_multiply", 'call': "*="},
                    "/=": {'name': "assign_divide", 'call': "/="},
                    "neg": {'name': "negative", 'call': "neg_one"},
                    "sin": {'name': "numpy_sin", 'call': "np.sin"},
                    "cos": {'name': "numpy_cos", 'call': "np.cos"},
                    "tan": {'name': "numpy_tan", 'call': "np.tan"},
                    "atan": {'name': "numpy_atan", 'call': "np.arctan"},
                    "abs": {'name': "numpy_abs", 'call': "np.abs"},
                    "sqrt": {'name': "numpy_sqrt", 'call': "np.sqrt"},
                    "sq": {'name': "numpy_square", 'call': "np.square"},
                    "exp": {'name': "numpy_exp", 'call': "np.exp"},
                    "max": {'name': "numpy_max", 'call': "np.maximum"},
                    "min": {'name': "numpy_min", 'call': "np.minimum"},
                    "argmax": {'name': "numpy_transpose", 'call': "np.argmax"},
                    "argmin": {'name': "numpy_argmin", 'call': "np.argmin"},
                    "round": {'name': "numpy_round", 'call': "np.round"},
                    "sum": {'name': "numpy_sum", 'call': "np.sum"},
                    "mean": {'name': "numpy_mean", 'call': "np.mean"},
                    "concat": {'name': "numpy_concatenate", 'call': "np.concatenate"},
                    "reshape": {'name': "numpy_reshape", 'call': "np.reshape"},
                    "append": {'name': "numpy_append", 'call': "np.append"},
                    "shape": {'name': "numpy_shape", 'call': "np.shape"},
                    "dtype": {'name': "numpy_dtype", 'call': "np.dtype"},
                    'squeeze': {'name': "numpy_squeeze", 'call': "np.squeeze"},
                    'expand': {'name': 'numpy_expand', 'call': "np.expand_dims"},
                    "roll": {'name': "numpy_roll", 'call': "np.roll"},
                    "cast": {'name': "numpy_cast", 'call': "np.asarray"},
                    "randn": {'name': "numpy_randn", 'call': "np.randn"},
                    "ones": {'name': "numpy_ones", 'call': "np.ones"},
                    "zeros": {'name': "numpy_zeros", 'call': "np.zeros"},
                    "range": {'name': "numpy_arange", 'call': "np.arange"},
                    "softmax": {'name': "pyrates_softmax", 'call': "pr_softmax"},
                    "sigmoid": {'name': "pyrates_sigmoid", 'call': "pr_sigmoid"},
                    "tanh": {'name': "numpy_tanh", 'call': "np.tanh"},
                    "index": {'name': "pyrates_index", 'call': "pyrates_index"},
                    "mask": {'name': "pyrates_mask", 'call': "pr_mask"},
                    "group": {'name': "pyrates_group", 'call': "pr_group"},
                    "asarray": {'name': "numpy_asarray", 'call': "np.asarray"},
                    "no_op": {'name': "pyrates_identity", 'call': "pr_identity"},
                    "interpolate": {'name': "pyrates_interpolate", 'call': "pr_interp"},
                    "interpolate_1d": {'name': "pyrates_interpolate_1d", 'call': "pr_interp_1d"},
                    "interpolate_nd": {'name': "pyrates_interpolate_nd", 'call': "pr_interp_nd"},
                    }
        if ops:
            self.ops.update(ops)

        # base data-types
        self.dtypes = {"float16": np.float16,
                       "float32": np.float32,
                       "float64": np.float64,
                       "int16": np.int16,
                       "int32": np.int32,
                       "int64": np.int64,
                       "uint16": np.uint16,
                       "uint32": np.uint32,
                       "uint64": np.uint64,
                       "complex64": np.complex64,
                       "complex128": np.complex128,
                       "bool": np.bool
                       }
        if dtypes:
            self.dtypes.update(dtypes)

        # further attributes
        self.vars = dict()
        self.layers = [[]]
        self.state_vars = []
        self.lhs_vars = []
        self.var_counter = {}
        self.op_counter = {}
        self.layer = 0
        self.op_indices = {}
        self._float_def = self.dtypes[float_default_type]
        self.name = name
        self._base_layer = 0
        self._input_layer_added = False
        self._imports = ["import numpy as np", "from pyrates.backend.funcs import *"]
        if imports:
            for imp in imports:
                if imp not in self._imports:
                    self._imports.append(imp)
        self._input_names = []

        # create build dir
        self._orig_dir = os.getcwd()
        if build_dir:
            os.makedirs(build_dir, exist_ok=True)
        dir_name = f"{build_dir}/pyrates_build" if build_dir else "pyrates_build"
        try:
            os.mkdir(dir_name)
        except FileExistsError:
            pass
        except FileNotFoundError as e:
            # for debugging
            raise e
        self._build_dir = f"{dir_name}/{self.name}"
        try:
            os.mkdir(self._build_dir)
        except FileExistsError:
            rmtree(self._build_dir)
            os.mkdir(self._build_dir)
        sys.path.append(self._build_dir)

    def run(self,
            T: float,
            dt: float,
            outputs: Optional[dict] = None,
            inputs: Optional[list] = None,
            dts: Optional[int] = None,
            solver: str = 'euler',
            out_dir: Optional[str] = None,
            profile: bool = False,
            verbose: bool = True,
            **kwargs
            ) -> tuple:
        """Executes all operations in the backend graph for a given number of steps.

        Parameters
        ----------
        T
            Simulation time.
        dt
            Simulation step size.
        outputs
            Variables in the graph to store the history from.
        inputs
            Extrinsic, time-dependent inputs to graph input variables.
        dts
            Sampling step-size.
        solver
            Type of the numerical solver to use.
        out_dir
            Directory to write the session log into.
        profile
            If true, the total graph execution time will be printed and returned.
        verbose
            If true, updates about the simulation process will be displayed in the terminal.

        Returns
        -------
        Union[Tuple[Dict[str, tf.Variable]], Tuple[dict, float]]
            If `profile` was requested, a tuple is returned that contains
                1) the results dictionary
                2) the simulation time.
            If not, a tuple containing only the results dictionary is returned which contains a numpy array with results
             for each output key that was provided via `outputs`. This allows to call the run functioon like this:
             `output, *time = run(*args, **kwargs)`

        """

        # preparations
        ##############

        if not dts:
            dts = dt

        # initialize session log
        if out_dir:
            # TODO : implement log files
            pass

        # initialize profiler
        if profile:
            t0 = time.time()

        # add inputs to graph
        continuous = 'scipy' in solver
        t = self.add_input_layer(inputs=inputs, T=T, continuous=continuous)

        if verbose:
            print("    ...user-defined inputs have been added to the model.")

        # graph execution
        #################

        # map layers that need to be executed to compiled network structure
        decorator = kwargs.pop('decorator', None)
        decorator_kwargs = kwargs.pop('decorator_kwargs', {})
        rhs_func, args, state_vars, var_map = self.compile(self._build_dir, decorator=decorator, **decorator_kwargs)

        if verbose:
            print("    ...the run function has been compiled.")

        # create output indices
        output_indices = []
        if outputs:
            for out_key, out_vars in outputs.items():
                for n, (idx, _) in enumerate(out_vars):
                    output_indices.append([i-self.idx_start for i in idx] if type(idx) is list else idx-self.idx_start)
                    outputs[out_key][n][0] = len(output_indices)-1
        else:
            for i in range(len(self.state_vars)):
                output_indices.append(i)
                out_key = self.state_vars[i]
                if len(self.vars[out_key].shape) > 0:
                    outputs[out_key] = [(i, [f"{out_key}_{j}" for j in range(self.vars[out_key].shape[0])])]
                else:
                    outputs[out_key] = [(i, [out_key])]

        # simulate backend behavior for each time-step
        func_args = self._process_func_args(args, var_map, dt)

        if verbose:
            print("starting the simulation.")

        times, results = self._solve(rhs_func=rhs_func, func_args=func_args, T=T, dt=dt, dts=dts, t=t, solver=solver,
                                     output_indices=output_indices, **kwargs)

        if verbose:
            print("Simulation finished!\n")

        # output storage and clean-up
        #############################

        # store output variables in output dictionary
        for i, (out_key, out_vars) in enumerate(outputs.items()):
            node_col = []
            for _, node_keys in out_vars:
                node_col += node_keys
            outputs[out_key] = (np.asarray(results[i]), node_col)

        # store profiling results
        if profile:
            sim_time = time.time() - t0
            return outputs, times, sim_time

        return outputs, times

    def add_var(self,
                vtype: str,
                name: Optional[str] = None,
                value: Optional[Any] = None,
                shape: Optional[Union[tuple, list, np.shape]] = None,
                dtype: Optional[Union[str, np.dtype]] = None,
                **kwargs
                ) -> NumpyVar:
        """Adds a variable to the backend.

        Parameters
        ----------
        vtype
            Variable type. Can be
                - `state_var` for variables that can change over time.
                - `constant` for non-changing variables.
        name
            Name of the variable.
        value
            Value of the variable. Not needed for placeholders.
        shape
            Shape of the variable.
        dtype
            Datatype of the variable.
        kwargs
            Additional keyword arguments.

        Returns
        -------
        PyRatesVar
            Handle for the numpy variable.

        """

        # extract variable scope
        scope = kwargs.pop('scope', None)
        if scope:
            name = f'{scope}/{name}'

        if name in self.vars:
            return self.get_var(name)

        # create variable
        var, name = self._create_var(vtype=vtype, dtype=dtype, shape=shape, value=value, name=name,
                                     squeeze=kwargs.pop('squeeze', True))

        # ensure uniqueness of variable names
        if var.short_name in self.var_counter and name not in self.vars:
            rename = True if var.short_name not in ["y",  "y_delta", "t",  "times"] else False
            if var.vtype == 'constant' and rename:
                for var_tmp in self.vars.values():
                    rename = not self._compare_vars(var, var_tmp)
                    if rename:
                        break
            if rename:
                name_old = var.short_name
                name_new = f"{name_old}_{self.var_counter[name_old]}"
                var.short_name = name_new
                self.var_counter[name_old] += 1
        else:
            self.var_counter[var.short_name] = 0

        self.vars[name] = var
        return var

    def remove_var(self, name: str) -> Union[NumpyVar, None]:
        """Removes variable from backend and returns it.

        Parameters
        ----------
        name
            Variable name.

        Returns
        -------
        NumpyVar

        """
        if name in self.var_counter:
            self.var_counter[name] -= 1
            if self.var_counter[name] < 0:
                self.var_counter.pop(name)
        if name in self.vars:
            return self.vars.pop(name)
        else:
            return None

    def add_op(self,
               op_name: str,
               *args,
               **kwargs
               ) -> Union[PyRatesOp, NumpyVar]:
        """Add operation to the backend.

        Parameters
        ----------
        op_name
            Key of the operation. Needs to be a key of `backend.ops`.
        args
            Positional arguments to be passed to the operation.
        kwargs
            Keyword arguments to be passed to the operation, except for scope and dependencies, which are
            extracted before.

        Returns
        -------
        Union[PyRatesOp, PyRatesVar]
            Handle for the lambda-numpy function

        """

        # process input arguments
        #########################

        # extract scope
        scope = kwargs.pop('scope', None)

        # generate operator identifier
        name = kwargs.pop('name', None)
        if name and scope:
            name = f'{scope}/{name}'
        elif scope:
            name = f'{scope}/assign' if '=' in op_name else f"{scope}/{self.ops[op_name]['name']}"
        else:
            name = f'assign' if '=' in op_name else f"{self.ops[op_name]['name']}"
        if name in self.op_counter:
            name_old = name
            name = f"{name}_{self.op_counter[name]}"
            self.op_counter[name_old] += 1
        else:
            self.op_counter[name] = 0

        # extract and process graph dependencies
        dependencies = kwargs.pop('dependencies', [])
        if dependencies:
            found = False
            for dep in dependencies:
                if dep.name == name:
                    found = True
                    break
            if found:
                self.add_layer()

        # create operation
        ##################

        try:

            # try generating the operator
            op = self._create_op(op_name, name, *args)

            # make sure that the output shape of the operation matches the expectations
            in_shape = []
            expand_ops = ('@', 'index', 'concat', 'expand', 'stack', 'group', 'asarray', 'interpolate', 'interpolate_nd'
                          )
            if op_name not in expand_ops:
                for arg in args:
                    if hasattr(arg, 'shape'):
                        in_shape.append(sum(tuple(arg.shape)))
                    elif type(arg) in (tuple, list):
                        if hasattr(arg[0], 'shape'):
                            in_shape.append(sum(tuple(arg[0].shape)))
                        else:
                            in_shape.append(sum(arg))
                    else:
                        in_shape.append(1)
                if hasattr(op, 'shape') and (sum(tuple(op.shape)) > max(in_shape)):
                    arg1, arg2 = self.broadcast(args[0], args[1])
                    op = self._create_op(op_name, name, arg1, arg2, *args[2:])

        except Exception:

            if type(args[0]) in (tuple, list) and len(args[0]) > 1:

                # broadcast entries of argument tuples
                args_tmp = self.broadcast(args[0][0], args[0][1])
                args_tmp = (args_tmp[0], args_tmp[1]) + tuple(args[0][2:])
                args_new = args_tmp + args[1:]

            elif '=' in op_name and len(args) > 2:

                # for assign operations with index, apply index temporarily and broadcast rhs to match indexed lhs
                arg1_tmp = self._create_op('index', 'test_idx', args[0], args[2])
                args_tmp = self.broadcast(arg1_tmp, args[1])
                args_new = (args[0], args_tmp[1], args[2])

            else:

                # broadcast leading 2 arguments to operation
                args_tmp = self.broadcast(args[0], args[1])
                args_new = args_tmp + args[2:]

            # try to generate operation again
            op = self._create_op(op_name, name, *args_new)

        # remove op inputs from layers (need not be evaluated anymore)
        for op_name_tmp in op.input_ops:
            if 'assign' not in op_name_tmp:
                idx_1, idx_2 = self.op_indices[op_name_tmp]
                self._set_op(None, idx_1, idx_2)

        # for constant ops, add a constant to the graph, containing the op evaluation
        if op.is_constant and op_name != 'no_op':
            new_var = op.numpy()
            if hasattr(new_var, 'shape'):
                name = f'{name}_evaluated'
                return self.add_var(vtype='constant', name=name, value=new_var)
            else:
                return new_var

        # add op to the graph
        self._add_op(op)
        self.op_indices[op.name] = (self.layer, len(self.get_current_layer())-1)
        return op

    def add_layer(self, to_beginning=False) -> None:
        """Adds a new layer to the backend and sets cursor to that layer.

        Parameters
        ----------
        to_beginning
            If true, layer is appended to the beginning of the layer stack instead of the end

        Returns
        -------
        None

        """

        if to_beginning:
            self.layers = [[]] + self.layers
            self._base_layer += 1
            self.layer = -self._base_layer
        else:
            self.layer = len(self.layers) - self._base_layer
            self.layers.append([])

    def add_output_layer(self, outputs, sampling_steps, out_shapes) -> dict:
        """

        Parameters
        ----------
        outputs
        sampling_steps
        out_shapes

        Returns
        -------

        """

        output_col = {}

        # add output storage layer to the graph
        if self._output_layer_added:

            out_idx = self.get_var("output_collection/out_var_idx")

            # jump to output layer
            self.top_layer()

        else:

            # add output layer
            self.add_layer()
            self._output_layer_added = True

            # create counting index for collector variables
            out_idx = self.add_var(vtype='state_var', name='out_var_idx', dtype='int32', shape=(1,), value=0,
                                   scope="output_collection")

            # create increment operator for counting index
            self.next_layer()
            self.add_op('+=', out_idx, np.ones((1,), dtype='int32'), scope="output_collection")
            self.previous_layer()

        # add collector variables to the graph
        for i, (var_col) in enumerate(outputs):

            shape = (sampling_steps + 1, len(var_col)) + out_shapes[i]
            key = f"output_col_{i}"
            output_col[key] = self.add_var(vtype='state_var', name=f"out_col_{i}", scope="output_collection",
                                           value=np.zeros(shape, dtype=self._float_def))
            var_stack = self.stack_vars(var_col)

            # add collect operation to the graph
            self.add_op('=', output_col[key], var_stack, out_idx, scope="output_collection")

        return output_col

    def add_input_layer(self, inputs: list, T: float, continuous=True) -> NumpyVar:
        """

        Parameters
        ----------
        inputs
        T
        continuous

        Returns
        -------

        """

        # add inputs to graph
        if self._input_layer_added:
            self.bottom_layer()
        else:
            self.add_layer(to_beginning=True)

        # create time-vector
        t = self.add_var('state_var', name='t', value=0.0, dtype=self._float_def, shape=())
        self.lhs_vars.append(t.short_name)

        if inputs:

            if continuous:

                from scipy.interpolate import interp1d
                time = np.linspace(0, T, inputs[0][0].shape[0])

                for (inp, target_var, idx) in inputs:

                    # create unique name of input variable
                    in_name_tmp = f"{target_var.short_name}_inp"
                    counter = 0
                    in_name = in_name_tmp
                    while in_name in self._input_names:
                        in_name = f"{in_name_tmp}_{counter}"
                        counter += 1
                    self._input_names.append(in_name)

                    # create interpolation operator
                    if len(inp.shape) > 1:
                        inp = inp.squeeze()
                    f = interp1d(time, inp, axis=0, copy=False, kind='linear')
                    f.shape = inp.shape[1:]
                    f = self.add_var(vtype='state_var', name=f"network_inputs/{in_name}", value=f)
                    in_var_interp = self.add_op('interpolate', f, t, scope="network_inputs")

                    # apply input to target variable
                    if idx:
                        self.add_op('=', target_var, in_var_interp, idx, scope="network_inputs")
                    else:
                        self.add_op('=', target_var, in_var_interp, scope="network_inputs")

            else:

                # create counting index for input variables
                time_step_idx = self.add_var(vtype='state_var', name='in_var_idx', dtype='int32', shape=(1,), value=0,
                                             scope="network_inputs")

                for (inp, target_var, idx) in inputs:

                    # create unique name of input variable
                    in_name_tmp = f"{target_var.short_name}_inp"
                    counter = 0
                    in_name = in_name_tmp
                    while in_name_tmp in self._input_names:
                        in_name = f"{in_name_tmp}_{counter}"
                        counter += 1
                    self._input_names.append(in_name)

                    # create time indexing operator
                    in_var = self.add_var(vtype='state_var', name=f"network_inputs/{in_name}", scope="network_inputs",
                                          value=inp)
                    in_var_indexed = self.add_op('index', in_var, time_step_idx, scope="network_inputs")

                    # apply input to target variable
                    if idx:
                        self.add_op('=', target_var, in_var_indexed, idx, scope="network_inputs")
                    else:
                        self.add_op('=', target_var, in_var_indexed, scope="network_inputs")

                # create increment operator for counting index
                time_step = self.add_var('constant', name='time_step_increment', value=np.ones((1,), dtype='int32'),
                                         scope="network_inputs")
                self.add_op('+=', time_step_idx, time_step, scope="network_inputs")

        return t

    def next_layer(self) -> None:
        """Jump to next layer in stack. If we are already at end of layer stack, add new layer to the stack and jump to
        that.
        """
        if self._base_layer+self.layer == len(self.layers)-1:
            self.add_layer()
        else:
            self.layer += 1

    def previous_layer(self) -> None:
        """Jump to previous layer in stack. If we are already at beginning of layer stack, add new layer to beginning of
        the stack and jump to that.
        """
        if self.layer == -self._base_layer:
            self.add_layer(to_beginning=True)
        else:
            self.layer -= 1

    def goto_layer(self, idx: int) -> None:
        """Jump to layer indicated by index.

        Parameters
        ----------
        idx
            Position of layer to jump towards.


        Returns
        -------
        None

        """
        self.layer = idx

    def remove_layer(self, idx) -> None:
        """Removes layer at index from stack.

        Parameters
        ----------
        idx
            Position of layer in graph.

        """
        self.layers.pop(self._base_layer + idx)
        if idx <= self._base_layer:
            self._base_layer -= 1

    def top_layer(self) -> int:
        """Jump to top layer of the stack.
        """
        self.layer = len(self.layers)-1-self._base_layer
        return self.layer

    def bottom_layer(self) -> int:
        """Jump to bottom layer of the stack.
        """
        self.layer = -self._base_layer
        return self.layer

    def get_current_layer(self) -> list:
        """Get operations in current layer.
        """
        return self.layers[self._base_layer + self.layer]

    def clear(self) -> None:
        """Removes all layers, variables and operations from graph. Deletes build directory.
        """
        self.vars.clear()
        self.layers = [[]]
        self.op_counter = 0
        self.var_counter = 0
        self.layer = 0
        rmtree(f"{self._orig_dir}/{self._build_dir}")
        if 'rhs_func' in sys.modules:
            del sys.modules['rhs_func']
        return self._orig_dir

    def get_layer(self, idx) -> list:
        """Retrieve layer from graph.

        Parameters
        ----------
        idx
            Position of layer in stack.

        """
        return self.layers[self._base_layer + idx]

    def get_var(self, name: str) -> Union[NumpyVar, PyRatesOp]:
        """Retrieve variable from graph.

        Parameters
        ----------
        name
            Identifier of the variable.

        Returns
        -------
        NumpyVar
            Variable from graph.

        """
        return self.vars[name]

    def eval_var(self, var: str) -> np.ndarray:
        """Get value of variable.

        Parameters
        ----------
        var
            Identifier of variable in graph.

        Returns
        -------
        np.ndarray
            Current value of the variable.

        """
        return self.vars[var].numpy()

    def compile(self, build_dir: Optional[str] = None, decorator: Optional[Callable] = None, **kwargs) -> tuple:
        """Compile the graph layers/operations. Creates python files containing the functions in each layer.

        Parameters
        ----------
        build_dir
            Directory in which to create the file structure for the simulation.
        decorator
            Decorator function that should be applied to the right-hand side evaluation function.
        kwargs
            decorator keyword arguments

        Returns
        -------
        tuple
            Contains tuples of layer run functions and their respective arguments.

        """

        # preparations
        ##############

        # remove empty layers and operators
        new_layer_idx = 0
        for layer_idx, layer in enumerate(self.layers.copy()):
            for op in layer.copy():
                if op is None:
                    layer.pop(layer.index(op))
            if len(layer) == 0:
                self.layers.pop(new_layer_idx)
            else:
                new_layer_idx += 1

        # remove previously imported rhs_funcs from system
        if 'rhs_func' in sys.modules:
            del sys.modules['rhs_func']

        # collect state variable and parameter vectors
        state_vars, params, var_map = self._process_vars()

        # update state variable getter operations to include the full state variable vector as first argument
        y = self.get_var('y')
        for key, (vtype, idx) in var_map.items():
            var = self.get_var(key)
            if vtype == 'state_var' and var.short_name != 'y':
                var.args[0] = y
                var.build_op(var.args)

        # create rhs evaluation function
        ################################

        # set up file header
        func_gen = CodeGen()
        for import_line in self._imports:
            func_gen.add_code_line(import_line)
            func_gen.add_linebreak()
        func_gen.add_linebreak()

        # define function head
        func_gen.add_code_line("def rhs_eval(t, y, params):")
        func_gen.add_linebreak()
        func_gen.add_indent()
        func_gen.add_linebreak()

        # declare constants
        args = [None for _ in range(len(params))]
        func_gen.add_code_line("# declare constants")
        func_gen.add_linebreak()
        updates, indices = [], []
        i = 0
        for key, (vtype, idx) in var_map.items():
            if vtype == 'constant':
                var = params[idx][1]
                func_gen.add_code_line(f"{var.short_name} = params[{idx}]")
                func_gen.add_linebreak()
                args[idx] = var
                if var.short_name != "y_delta":
                    updates.append(f"{var.short_name}")
                    indices.append(i)
                    i += 1
        func_gen.add_linebreak()

        # extract state variables from input vector y
        func_gen.add_code_line("# extract state variables from input vector")
        func_gen.add_linebreak()
        for key, (vtype, idx) in var_map.items():
            var = self.get_var(key)
            if vtype == 'state_var':
                func_gen.add_code_line(f"{var.short_name} = {var.value}")
                func_gen.add_linebreak()
        func_gen.add_linebreak()

        # add equations
        func_gen.add_code_line("# calculate right-hand side update of equation system")
        func_gen.add_linebreak()
        arg_updates = []
        for i, layer in enumerate(self.layers):
            for j, op in enumerate(layer):
                lhs = extract_lhs_var(op.value)
                find_arg = [arg == lhs for arg in updates]
                if any(find_arg):
                    idx = find_arg.index(True)
                    arg_updates.append((updates[idx], indices[idx]))
                func_gen.add_code_line(op.value)
                func_gen.add_linebreak()
        func_gen.add_linebreak()

        # update parameters where necessary
        func_gen.add_code_line("# update system parameters")
        func_gen.add_linebreak()
        for upd, idx in arg_updates:
            update_str = f"params[{idx}] = {upd}"
            if f"    {update_str}" not in func_gen.code:
                func_gen.add_code_line(update_str)
                func_gen.add_linebreak()

        # add return line
        func_gen.add_code_line(f"return {self.vars['y_delta'].short_name}")
        func_gen.add_linebreak()

        # save rhs function to file
        fname = f'{self._build_dir}/rhs_func'
        with open(f'{fname}.py', 'w') as f:
            f.writelines(func_gen.code)
            f.close()

        # import function from file
        exec(f"from rhs_func import rhs_eval", globals())
        rhs_eval = globals().pop('rhs_eval')

        # apply function decorator
        if decorator:
            rhs_eval = decorator(rhs_eval, **kwargs)

        return rhs_eval, args, state_vars, var_map

    def broadcast(self, op1: Any, op2: Any, **kwargs) -> tuple:
        """Tries to match the shapes of op1 and op2 such that op can be applied.

        Parameters
        ----------
        op1
            First argument to the operation.
        op2
            Second argument to the operation.
        kwargs
            Additional keyword arguments to be passed to the backend.

        Returns
        -------
        tuple
            Broadcasted op1 and op2.

        """

        # try to match shapes
        if not self._compare_shapes(op1, op2):

            # try adjusting op2 to match shape of op1
            adjust_first = True if hasattr(op2, 'short_name') else False
            op1_tmp, op2_tmp = self._match_shapes(op1, op2, adjust_second=adjust_first)

            if not self._compare_shapes(op1_tmp, op2_tmp):

                # try adjusting op1 to match shape of op2
                op1_tmp, op2_tmp = self._match_shapes(op1_tmp, op2_tmp, adjust_second=not adjust_first)

            if self._compare_shapes(op1_tmp, op2_tmp):
                op1, op2 = op1_tmp, op2_tmp

        return op1, op2

    def apply_idx(self, var: Any, idx: Any, update: Optional[Any] = None, update_type: str = None, *args) -> Any:
        """Applies index to a variable. IF update is passed, variable is updated at positions indicated by index.

        Parameters
        ----------
        var
            Variable to index/update
        idx
            Index to variable
        update
            Update to variable entries
        update_type
            Type of lhs update (e.g. `=` or `+=`)

        Returns
        -------
        Any
            Updated/indexed variable.

        """
        if update is None:
            return self.add_op('index', var, idx, *args)
        if not update_type:
            update_type = '='
        return self.add_op(update_type, var, update, idx, *args)

    def stack_vars(self, vars: tuple, **kwargs) -> np.ndarray:
        """Group variables into a stack, with the number of variables being the first dimension of the stack.

        Parameters
        ----------
        vars
            Variables to stack together. Should have the same shapes.
        kwargs
            Additional keyword arguments (Ignored in this version of the backend).

        Returns
        -------
        np.ndarray
            Stacked variables.

        """
        return self.add_op('asarray', vars)

    def to_file(self, fn: str):
        """Writes backend to file

        Parameters
        ----------
        fn

        Returns
        -------

        """

        pass

    def from_file(self, fn: str):
        """

        Parameters
        ----------
        fn

        Returns
        -------

        """

        pass

    @staticmethod
    def eval(ops: list) -> list:
        """Evaluates each operation in list.

        Parameters
        ----------
        ops
            List of operations to evaluate.

        """
        return [op.numpy() for op in ops]

    @staticmethod
    def eval_layer(layer: tuple) -> Any:
        """Evaluate layer run function (can only be used after compilation has been performed).

        Parameters
        ----------
        layer
            Layer tuple as returned by `compile`.

        Returns
        -------
        Any
            Result of layer evaluation.

        """
        return layer[0](*layer[1])

    def _set_op(self, op, idx1, idx2):
        self.layers[self._base_layer+idx1][idx2] = op

    def _add_op(self, op, layer=None):
        if layer is None:
            self.get_current_layer().append(op)
        else:
            self.get_layer(layer).append(op)

    def _solve(self, rhs_func, func_args, T, dt, dts, t, solver, output_indices, **kwargs):
        """

        Parameters
        ----------
        rhs_func
        func_args
        state_vars
        T
        dt
        dts
        t
        solver
        output_indices
        kwargs

        Returns
        -------

        """

        # choose solver
        ###############

        if solver == 'euler':

            times, results = self._integrate(rhs_func=rhs_func, func_args=func_args, T=T, dt=dt, dts=dts, t=t,
                                             output_indices=output_indices)

        elif solver == 'scipy':

            from scipy.integrate import solve_ivp

            # solve via scipy's ode integration function
            fun = lambda t, y: rhs_func(t, y, func_args)
            if dts:
                times = np.arange(0, T, dts)
                kwargs['t_eval'] = times
            outputs = solve_ivp(fun=fun, t_span=(float(t.numpy()), T), y0=self.vars['y'], first_step=dt,
                                **kwargs)
            results = [outputs['y'].T[:, idx] for idx in output_indices]
            times = outputs['t']
            self.vars['y'] = outputs['y'].T[:, -1]

        else:

            raise ValueError(f'Invalid solver type: {solver}. See docstring of `CircuitIR.run` for valid solver types.')

        return times, results

    def _integrate(self, rhs_func, func_args, T, dt, dts, t, output_indices):

        sampling_step = int(np.round(dts / dt, decimals=0))
        sampling_steps = int(np.round(T / dts, decimals=0))
        steps = int(np.round(T / dt, decimals=0))

        # initialize results storage vectors
        results = []
        for i, idx in enumerate(output_indices):
            if type(idx) is tuple:
                var_dim = idx[1]-idx[0]
            elif type(idx) is list:
                if all(np.diff(idx, n=1) == 1):
                    output_indices[i] = (idx[0], idx[-1]+1)
                    var_dim = idx[-1]+1-idx[0]
                else:
                    var_dim = len(idx)
            else:
                var_dim = 1
            results.append(np.zeros((sampling_steps, var_dim)))

        # solve via pyrates internal explicit euler algorithm
        state_vars = self.vars['y']
        sampling_idx = 0
        for i in range(steps):
            deltas = rhs_func(t, state_vars, func_args)
            t += dt
            state_vars += dt * deltas
            if i % sampling_step == 0:
                for idx1, idx2 in enumerate(output_indices):
                    results[idx1][sampling_idx, :] = state_vars[idx2[0]:idx2[1]] if type(idx2) is tuple \
                        else state_vars[idx2]
                sampling_idx += 1

        self.vars['y'] = state_vars
        times = np.arange(0, T, dts)

        return times, results

    def _match_shapes(self, op1: Any, op2: Any, adjust_second: bool = True) -> tuple:
        """Re-shapes op1 and op2 such that they can be combined via mathematical operations.

        Parameters
        ----------
        op1
            First operator.
        op2
            Second operator.
        adjust_second
            If true, the second operator will be re-shaped according to the shape of the first operator. If false,
            it will be done the other way around.

        Returns
        -------
        tuple
            The re-shaped operators.

        """

        if not hasattr(op1, 'shape'):
            op1 = self.add_var(vtype='constant', value=op1, name='c')
        if not hasattr(op2, 'shape'):
            op2 = self.add_var(vtype='constant', value=op2, name='c')

        if adjust_second:
            op_adjust = op2
            op_target = op1
        else:
            op_adjust = op1
            op_target = op2

        if len(op_target.shape) > len(op_adjust.shape):

            # add axis to op_adjust
            target_shape = op_target.shape
            idx = list(target_shape).index(1) if 1 in op_target.shape else len(target_shape) - 1
            op_adjust = self.add_op('expand', op_adjust, idx)

        elif (len(op_adjust.shape) > len(op_target.shape) and 1 in op_adjust.shape) or \
                (len(op_target.shape) == 2 and len(op_adjust.shape) == 2 and op_target.shape[1] != op_adjust.shape[0]
                 and 1 in op_adjust.shape):

            # cut singleton dimension from op_adjust
            old_shape = list(op_adjust.shape)
            idx = ",".join(['0' if s == 1 else ':' for s in old_shape])
            op_adjust = self.add_op('index', op_adjust, idx)

        if adjust_second:
            return op_target, op_adjust
        return op_adjust, op_target

    def _match_dtypes(self, op1: Any, op2: Any) -> tuple:
        """Match data types of two operators/variables.
        """

        if issubclass(type(op1), PyRatesOp):

            if issubclass(type(op2), PyRatesOp):

                # cast both variables to lowest precision
                for acc in ["16", "32", "64", "128"]:
                    if acc in op2.dtype:
                        return self.add_op('cast', op1, op2.dtype), op2

            return op1, self.add_op('cast', op2, op1.dtype)

        elif issubclass(type(op2), PyRatesOp):

            return self.add_op('cast', op1, op2.dtype), op2

        elif hasattr(op1, 'numpy') or type(op1) is np.ndarray:

            # cast op2 to numpy dtype of op1
            return op1, self.add_op('cast', op2, op1.dtype)

        elif hasattr(op2, 'numpy') or type(op2) is np.ndarray:

            # cast op1 to numpy dtype of op2
            return self.add_op('cast', op1, op2.dtype), op2

        else:

            # cast op2 to dtype of op1 referred from its type string
            return op1, self.add_op('cast', op2, str(type(op1)))

    def _create_var(self, vtype, dtype, shape, value, name, squeeze=True):
        return NumpyVar(vtype=vtype, dtype=dtype, shape=shape, value=value, name=name, backend=self, squeeze=squeeze)

    def _create_op(self, op, name, *args):
        if not self.ops[op]['call']:
            raise NotImplementedError(f"The operator `{op}` is not implemented for this backend ({self.name}). "
                                      f"Please consider passing the required operation to the backend initialization "
                                      f"or choose another backend.")
        if op in ["=", "+=", "-=", "*=", "/="]:
            if len(args) > 2 and hasattr(args[2], 'shape') and len(args[2].shape) > 1 and \
                    'bool' not in str(type(args[2])):
                args_tmp = list(args)
                if not hasattr(args_tmp[2], 'short_name'):
                    idx = self.add_var(vtype='state_var', name='idx', value=args_tmp[2])
                else:
                    idx = args_tmp[2]
                var, upd, idx = self._process_update_args_old(args[0], args[1], idx)
                idx_str = ",".join([f"{idx.short_name}[:,{i}]" for i in range(idx.shape[1])])
                args = (var, upd, idx_str, idx)
            return PyRatesAssignOp(self.ops[op]['call'], self.ops[op]['name'], name, *args,
                                   idx_l=self.idx_l, idx_r=self.idx_r)
        elif op == "index":
            return PyRatesIndexOp(self.ops[op]['call'], self.ops[op]['name'], name, *args, idx_l=self.idx_l,
                                  idx_r=self.idx_r)
        else:
            if op == "cast":
                args = list(args)
                for dtype in self.dtypes:
                    if dtype in str(args[1]):
                        args[1] = f"np.{dtype}"
                        break
                args = tuple(args)
            return PyRatesOp(self.ops[op]['call'], self.ops[op]['name'], name, *args)

    def _process_update_args_old(self, var, update, idx):
        """Preprocesses the index and a variable update to match the variable shape.

        Parameters
        ----------
        var
        update
        idx

        Returns
        -------
        tuple
            Preprocessed index and re-shaped update.

        """

        # pre-process args
        ###################

        shape = var.shape

        # match shape of index and update to shape
        ##########################################

        if tuple(update.shape) and idx.shape[0] != update.shape[0]:
            if update.shape[0] == 1 and len(update.shape) == 1:
                idx = self.add_op('reshape', idx, tuple(idx.shape) + (1,))
            elif idx.shape[0] == 1:
                update = self.add_op('reshape', update, (1,) + tuple(update.shape))
            elif len(update.shape) > len(idx.shape) and 1 in update.shape:
                singleton = list(update.shape).index(1)
                update = self.add_op('squeeze', update, singleton)
            else:
                raise ValueError(f'Invalid indexing. Operation of shape {shape} cannot be updated with updates of '
                                 f'shapes {update.shape} at locations indicated by indices of shape {idx.shape}.')
        elif not tuple(update.shape) and tuple(idx.shape):
            update = self.add_op('reshape', update, (1,))

        if len(idx.shape) < 2:
            idx = self.add_op('reshape', idx, tuple(idx.shape) + (1,))

        shape_diff = len(shape) - idx.shape[1]
        if shape_diff < 0:
            raise ValueError(f'Invalid index shape. Operation has shape {shape}, but received an index of '
                             f'length {idx.shape[1]}')

        # manage shape of updates
        update_dim = 0
        if len(update.shape) > 1:
            update_dim = len(update.shape[1:])
            if update_dim > shape_diff:
                singleton = -1 if update.shape[-1] == 1 else list(update.shape).index(1)
                update = self.add_op('squeeze', update, singleton)
                if len(update.shape) > 1:
                    update_dim = len(update.shape[1:])

        # manage shape of idx
        shape_diff = int(shape_diff) - update_dim
        if shape_diff > 0:
            indices = []
            indices.append(idx)
            for i in range(shape_diff):
                indices.append(self.add_op('zeros', tuple(idx.shape), idx.dtype))
            idx = self.add_op('concat', indices, 1)

        return var, update, idx

    def _process_update_args(self, var, update, idx=None):
        """Preprocesses the index and a variable update to match the variable shape.

        Parameters
        ----------
        var
        update
        idx

        Returns
        -------
        tuple
            Preprocessed index and re-shaped update.

        """

        # pre-process args
        shape = var.shape

        # match shape of index and update to shape
        ##########################################

        if len(shape) == len(update.shape) and shape != update.shape:

            # make sure that update dims match the variable shape
            update = self.add_op("squeeze", update)

        if shape[0] == idx.shape[0] and len(shape) > 1 and sum(idx.shape) > 1:

            # make sure that index scatters into first dimension of variable
            idx = self.add_op("index", idx, ":, None")

        elif update.shape == shape[1:] or update.shape == shape[:-1]:

            # make sure that index and update scatter into the first dimension of update
            update = self.add_op('reshape', update, (1,) + tuple(update.shape))
            idx = idx[None, :]

        return var, update, idx

    def _process_vars(self):
        """

        Returns
        -------

        """
        state_vars, constants, var_map = [], [], {}
        s_idx, c_idx, s_len = 0, 0, 0
        for key, var in self.vars.items():
            key, state_var = self._is_state_var(key)
            if state_var:
                state_vars.append((key, var))
                var_map[key] = ('state_var', (s_idx, s_len))
                s_idx += 1
                s_len += 1
            elif var.vtype == 'constant' or (var.short_name not in self.lhs_vars and key != 'y'):
                constants.append((key, var))
                var_map[key] = ('constant', c_idx)
                c_idx += 1
        return state_vars, constants, var_map

    def _is_state_var(self, key):
        return key, key in self.state_vars

    @staticmethod
    def _process_func_args(args, var_map, dt):
        for key, var_info in var_map.items():
            if 'times' in key:
                args[var_info[1]] -= dt
        return list(args)

    @staticmethod
    def _compare_vars(v1, v2):
        if v1.short_name == v2.short_name and v1.shape == v2.shape:
            if sum(v1.shape) > 1:
                return all(v1.flatten() != v2.flatten())
            else:
                return v2 != v1

    @staticmethod
    def _compare_shapes(op1: Any, op2: Any, index=False) -> bool:
        """Checks whether the shapes of op1 and op2 are compatible with each other.

        Parameters
        ----------
        op1
            First operator.
        op2
            Second operator.

        Returns
        -------
        bool
            If true, the shapes of op1 and op2 are compatible.

        """

        if hasattr(op1, 'shape') and hasattr(op2, 'shape'):
            if op1.shape == op2.shape:
                return True
            elif len(op1.shape) > 1 and len(op2.shape) > 1:
                return True
            elif len(op1.shape) == 0 and len(op2.shape) == 0:
                return True
            else:
                return False
        else:
            return True

    @staticmethod
    def _deepcopy(x):
        return deepcopy(x)


class CodeGen:
    """Generates python code. Can add code lines, line-breaks, indents and remove indents.
    """

    def __init__(self):
        self.code = []
        self.lvl = 0

    def generate(self):
        """Generates a single code string from its history of code additions.
        """
        return ''.join(self.code)

    def add_code_line(self, code_str):
        """Add code line string to code.
        """
        self.code.append("\t" * self.lvl + code_str)

    def add_linebreak(self):
        """Add a line-break to the code.
        """
        self.code.append("\n")

    def add_indent(self):
        """Add an indent to the code.
        """
        self.lvl += 1

    def remove_indent(self):
        """Remove an indent to the code.
        """
        if self.lvl == 0:
            raise(SyntaxError("internal error in code generator"))
        self.lvl -= 1


def extract_lhs_var(eq_str: str) -> str:
    """

    Parameters
    ----------
    eq_str

    Returns
    -------

    """

    lhs = eq_str.split("=")[0]
    lhs = lhs.replace(" ", "")
    if "[" in lhs:
        idx = lhs.index("[")
        lhs = lhs[:idx]
    return lhs



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network models and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation
"""
Test suite to determine, whether caching of identical operator works properly.
"""

__author__ = "Daniel Rose"
__status__ = "Development"

import pytest


def setup_module():
    print("\n")
    print("================================")
    print("| Test Suite: Operator Caching |")
    print("================================")


def test_op_caching_nodes():
    """Test the case that two operator templates are identical up to variable definitions, where one inherits from the
    other"""
    from pyrates.frontend import CircuitTemplate

    circuit = CircuitTemplate.from_yaml("model_templates.test_resources.test_operator_caching_templates.TestCircuit1").apply()

    # access nodes
    node1 = circuit["node1"]
    node2 = circuit["node2"]

    # verify that operator labels and contents are identical
    op1, op1_dict = next(iter(node1.op_graph.nodes(data=True)))
    op2, op2_dict = next(iter(node2.op_graph.nodes(data=True)))

    assert op1 == op2
    assert op1_dict == op2_dict
    assert op1_dict["operator"] is op2_dict["operator"]

    # verify that values in nodes are indeed different (but refer to same operator)
    assert node1.values["TestOpLin0"]["c"] == 0
    assert node2.values["TestOpLin0"]["c"] == 1

    # now do the vectorization step
    circuit = circuit.optimize_graph_in_place()

# def test_op_caching_edges():
# still need to implement a test for edges, because they behave a little different.











# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural 
# network models and simulations. See also: 
# https://github.com/pyrates-neuroscience/PyRates
# 
# Copyright (C) 2017-2018 the original authors (Richard Gast and 
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain 
# Sciences ("MPI CBS") and contributors
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
# 
# CITATION:
# 
# Richard Gast and Daniel Rose et. al. in preparation
"""
"""

import pytest

__author__ = "Daniel Rose"
__status__ = "Development"


def test_ir_vectorization():
    """Test, if apply() functions all work properly"""

    path = "model_templates.jansen_rit.circuit.JansenRitCircuit"
    from pyrates.frontend.template.circuit import CircuitTemplate
    from pyrates.ir.circuit import CircuitIR

    template = CircuitTemplate.from_yaml(path)

    circuit = template.apply()  # type: CircuitIR
    circuit2 = circuit.optimize_graph_in_place()

    # these should actually be the same
    assert circuit is circuit2

    # ensure only vector nodes exist
    for node in circuit.nodes:
        assert node.startswith("vector_")

    # for source, target, key in circuit.edges:
    #     if "node" in source:
    #         assert "coupling" in target
    #     else:
    #         assert "coupling" in source and "node" in target

    # for source, target, data in circuit2.edges(data=True):
    #     # check that no operators are left in the edges of the rearranged circuit
    #     assert len(data["edge_ir"].op_graph) == 0

    # check that operator from previous edges is indeed in target nodes
    # original_edge = circuit.edges[(source, target, 0)]["edge_ir"]
    # original_op = list(original_edge.op_graph.nodes)[0]
    # assert f"{original_op}.0" in circuit2[target]


@pytest.mark.xfail
def test_ir_compilation():
    path = "model_templates.jansen_rit.circuit.JansenRitCircuit"
    from pyrates.ir.circuit import CircuitIR

    circuit = CircuitIR.from_yaml(path)

    circuit.compile()



# -*- coding: utf-8 -*-
#
#
# PyRates software framework for flexible implementation of neural
# network model_templates and simulations. See also:
# https://github.com/pyrates-neuroscience/PyRates
#
# Copyright (C) 2017-2018 the original authors (Richard Gast and
# Daniel Rose), the Max-Planck-Institute for Human Cognitive Brain
# Sciences ("MPI CBS") and contributors
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>
#
# CITATION:
#
# Richard Gast and Daniel Rose et. al. in preparation

"""This module provides parser classes and functions to parse string-based equations into symbolic representations of
operations.
"""

# external imports
import math
import typing as tp
from numbers import Number
from pyparsing import Literal, CaselessLiteral, Word, Combine, Optional, \
    ZeroOrMore, Forward, nums, alphas, ParserElement

# meta infos
__author__ = "Richard Gast"
__status__ = "development"


# expression parsers (lhs/rhs of an equation)
#############################################


class ExpressionParser(ParserElement):
    """Base class for parsing mathematical expressions from a string format into a symbolic representation of the
    mathematical operation expressed by it.

    Parameters
    ----------
    expr_str
        Mathematical expression in string format.
    args
        Dictionary containing all variables and functions needed to evaluate the expression.
    backend
        Backend instance in which to parse all variables and operations.
        See `pyrates.backend.numpy_backend.NumpyBackend` for a full documentation of the backends methods and
        attributes.
    kwargs
        Additional keyword arguments to be passed to the backend functions.

    Attributes
    ----------
    lhs
        Boolean, indicates whether expression is left-hand side or right-hand side of an equation
    rhs
        PyRatesOp for the evaluation of the right-hand side of the equation
    args
        Dictionary containing the variables of an expression
    solve
        Only relevant for lhs expressions. If True, lhs will be treated as a first-order ordinary differential equation.
    expr_str
        String representation of the mathematical expression
    expr
        Symbolic (Pyparsing-based) representation of mathematical expression.
    expr_stack
        List representation of the syntax tree of the (parsed) mathematical expression.
    expr_list
        List representation of the mathematical expression.
    op
        Operator for calculating the mathematical expression (symbolic representation).
    _op_tmp
        Helper variable for building `op`.
    ops
        Dictionary containing all mathematical operations that are allowed for a specific instance of the
        `ExpressionParser` (e.g. +, -, *, /).
    funcs
        Dictionary containing all additional functions that can be used within mathematical expressions with a specific
        instance of the `ExpressionParser` (e.g. sum(), reshape(), float32()).
    dtypes
        Dictionary containing all data-types that can be used within mathematical expressions with a specific instance
        of the `ExpressionParser` (e.g. float32, bool, int32).

    """

    def __init__(self, expr_str: str, args: dict, backend: tp.Any, **kwargs) -> None:
        """Instantiates expression parser.
        """

        # call super init
        #################

        super().__init__()

        # bind attributes to instance
        #############################

        # input arguments
        self.vars = args.copy()
        self.var_map = {}
        self.backend = backend
        self.parser_kwargs = kwargs

        self.lhs, self.rhs, self._diff_eq, self._assign_type, self.lhs_key = self._preprocess_expr_str(expr_str)

        # add functions from args dictionary to backend, if passed
        for key, val in args.items():
            if callable(val):
                self.backend.ops[key] = val

        # additional attributes
        self.expr_str = expr_str
        self.expr = None
        self.expr_stack = []
        self.expr_list = []
        self.op = None
        self._finished_rhs = False
        self._instantaneous = kwargs.pop('instantaneous', False)

        # define algebra
        ################

        if not self.expr:

            # general symbols
            point = Literal(".")
            comma = Literal(",")
            colon = Literal(":")
            e = CaselessLiteral("E")
            pi = CaselessLiteral("PI")

            # parentheses
            par_l = Literal("(")
            par_r = Literal(")").setParseAction(self._push_first)
            idx_l = Literal("[")
            idx_r = Literal("]")

            # basic mathematical operations
            plus = Literal("+")
            minus = Literal("-")
            mult = Literal("*")
            div = Literal("/")
            mod = Literal("%")
            dot = Literal("@")
            exp_1 = Literal("^")
            exp_2 = Combine(mult + mult)
            transp = Combine(point + Literal("T"))
            inv = Combine(point + Literal("I"))

            # numeric types
            num_float = Combine(Word("-" + nums, nums) +
                                Optional(point + Optional(Word(nums))) +
                                Optional(e + Word("-" + nums, nums)))
            num_int = Word("-" + nums, nums)

            # variables and functions
            name = Word(alphas, alphas + nums + "_$")
            func_name = Combine(name + par_l, adjacent=True)

            # math operation groups
            op_add = plus | minus
            op_mult = mult | div | dot | mod
            op_exp = exp_1 | exp_2 | inv | transp

            # logical operations
            greater = Literal(">")
            less = Literal("<")
            equal = Combine(Literal("=") + Literal("="))
            unequal = Combine(Literal("!") + Literal("="))
            greater_equal = Combine(Literal(">") + Literal("="))
            less_equal = Combine(Literal("<") + Literal("="))

            # logical operations group
            op_logical = greater_equal | less_equal | unequal | equal | less | greater

            # pre-allocations
            self.expr = Forward()
            exponential = Forward()
            index_multiples = Forward()

            # basic organization units
            index_start = idx_l.setParseAction(self._push_first)
            index_end = idx_r.setParseAction(self._push_first)
            index_comb = colon.setParseAction(self._push_first)
            arg_comb = comma.setParseAction(self._push_first)
            arg_tuple = par_l + ZeroOrMore(self.expr.suppress() + Optional(arg_comb)) + par_r
            func_arg = arg_tuple | self.expr.suppress()

            # basic computation unit
            atom = (func_name + Optional(func_arg.suppress()) + ZeroOrMore(arg_comb.suppress() + func_arg.suppress()) +
                    par_r.suppress() | name | pi | e | num_float | num_int).setParseAction(self._push_neg_or_first) | \
                   (par_l.setParseAction(self._push_last) + self.expr.suppress() + par_r).setParseAction(self._push_neg)

            # apply indexing to atoms
            indexed = (Optional(minus) + atom).setParseAction(self._push_neg) + \
                      ZeroOrMore((index_start + index_multiples + index_end))
            index_base = (self.expr.suppress() | index_comb)
            index_full = index_base + ZeroOrMore((index_comb + index_base)) + ZeroOrMore(index_comb)
            index_multiples << index_full + ZeroOrMore((arg_comb + index_full))

            # hierarchical relationships between mathematical and logical operations
            boolean = indexed + Optional((op_logical + indexed).setParseAction(self._push_first))
            exponential << boolean + ZeroOrMore((op_exp + Optional(exponential)).setParseAction(self._push_first))
            factor = exponential + ZeroOrMore((op_mult + exponential).setParseAction(self._push_first))
            expr = factor + ZeroOrMore((op_add + factor).setParseAction(self._push_first))
            self.expr << expr #(Optional(minus) + expr).setParseAction(self._push_neg)

    def parse_expr(self) -> tuple:
        """Parses string-based mathematical expression/equation.

        Returns
        -------
        tuple
            left-hand side, right-hand side and variables of the parsed equation.
        """

        # extract symbols and operations from equations right-hand side
        self.expr_list = self.expr.parseString(self.rhs)
        self._check_parsed_expr(self.rhs)

        # parse rhs into backend
        self.rhs = self.parse(self.expr_stack[:])

        # post rhs parsing steps
        if hasattr(self.rhs, 'vtype') or "float" in str(type(self.rhs)) or "int" in str(type(self.rhs)):
            self.rhs = self.backend.add_op('no_op', self.rhs, **self.parser_kwargs)
        self.clear()
        self._finished_rhs = True

        # extract symbols and operations from left-hand side
        self.expr_list = self.expr.parseString(self.lhs)
        self._check_parsed_expr(self.lhs)

        # parse lhs into backend
        self._update_lhs()

        return self.lhs, self.rhs, self.vars

    def parse(self, expr_stack: list) -> tp.Any:
        """Parse elements in expression stack into the backend.

        Parameters
        ----------
        expr_stack
            Ordered list with expression variables and operations. Needs to be processed from last to first item.

        Returns
        -------
        tp.Any
            Parsed expression stack element (object type depends on the backend).

        """

        # get next operation from stack
        op = expr_stack.pop()

        # check type of operation
        #########################

        if op == '-one':

            # multiply expression by minus one
            self.op = self.backend.add_op('*', self.parse(expr_stack), -1, **self.parser_kwargs)

        elif op in ["*=", "/=", "+=", "-=", "="]:

            # collect rhs
            op1 = self.parse(expr_stack)

            # collect lhs
            indexed_lhs = True if "]" in expr_stack else False
            op2 = self.parse(expr_stack)

            # combine elements via mathematical/boolean operator
            if indexed_lhs:
                self.op = self._apply_idx(op=op2[0], idx=op2[1], update=op1, update_type=op, **self.parser_kwargs)
            else:
                self.op = self.backend.add_op(op, op2, op1, **self.parser_kwargs)

        elif op in "+-/**^@<=>=!==%":

            # collect elements to combine
            op2 = self.parse(expr_stack)
            op1 = self.parse(expr_stack)

            # combine elements via mathematical/boolean operator
            self.op = self.backend.add_op(op, op1, op2, **self.parser_kwargs)

        elif ".T" == op or ".I" == op:

            # transpose/invert expression
            self.op = self.backend.add_op(op, self.parse(expr_stack), **self.parser_kwargs)

        elif op == "]":

            # parse indices
            indices = []
            while len(expr_stack) > 0 and expr_stack[-1] != "[":
                index = []
                while len(expr_stack) > 0 and expr_stack[-1] not in ",[":
                    if expr_stack[-1] == ":":
                        index.append(expr_stack.pop())
                    else:
                        try:
                            int(expr_stack[-1])
                            index.append(expr_stack.pop())
                        except ValueError:
                            tmp = self._finished_rhs
                            self._finished_rhs = False
                            index.append(self.parse(expr_stack))
                            self._finished_rhs = tmp
                indices.append(index[::-1])
                if expr_stack[-1] == ",":
                    expr_stack.pop()
            expr_stack.pop()

            # build string-based representation of idx
            if 'idx' not in self.vars.keys():
                self.vars['idx'] = {}
            idx = ""
            i = 0
            for index in indices[::-1]:
                for ind in index:
                    if type(ind) == str:
                        idx += ind
                    elif isinstance(ind, Number):
                        idx += f"{ind}"
                    else:
                        self.vars['idx'][f'idx_var_{i}'] = ind
                        idx += f"idx_var_{i}"
                    i += 1
                idx += ","
            idx = idx[0:-1]

            # extract variable and apply idx if its a rhs variable. Else return variable and index
            if self._finished_rhs:
                op = expr_stack.pop(-1)
                if op in self.vars:
                    op_to_idx = self.vars[op]
                else:
                    op_to_idx = self.parse([op])
                self.op = (op_to_idx, idx)
            else:
                op_to_idx = self.parse(expr_stack)
                op_idx = self._apply_idx(op_to_idx, idx, **self.parser_kwargs)
                self.op = op_idx

        elif op == "PI":

            # return float representation of pi
            self.op = math.pi

        elif op == "E":

            # return float representation of e
            self.op = math.e

        elif op in self.vars:

            # extract constant/variable from args dict
            self.op = self.vars[op]

        elif op[-1] == "(":

            expr_stack.pop(-1)

            # parse arguments
            args = []
            while len(expr_stack) > 0:
                args.append(self.parse(expr_stack))
                if len(expr_stack) == 0 or expr_stack[-1] != ",":
                    break
                else:
                    expr_stack.pop()

            # apply function to arguments
            try:
                if len(args) == 1:
                    self.op = self.backend.add_op(op[0:-1], args[0], **self.parser_kwargs)
                else:
                    self.op = self.backend.add_op(op[0:-1], *tuple(args[::-1]), **self.parser_kwargs)
            except KeyError:
                if any(["float" in op, "bool" in op, "int" in op, "complex" in op]):
                    self.op = self.backend.add_op('cast', args[0], op[0:-1], **self.parser_kwargs)
                else:
                    raise KeyError(f"Undefined function in expression: {self.expr_str}. {op[0:-1]} needs to be "
                                   f"provided in arguments dictionary.")

        elif op == ")":

            # check whether expression in parenthesis is a group of arguments to a function
            start_par = -1
            found_end = 0
            while found_end < 1:
                if "(" in expr_stack[start_par]:
                    found_end += 1
                if ")" in expr_stack[start_par]:
                    found_end -= 1
                start_par -= 1
            if "," in expr_stack[start_par+1:]:

                args = []
                while True:
                    args.append(self.parse(expr_stack))
                    if expr_stack[-1] == ",":
                        expr_stack.pop(-1)
                    elif expr_stack[-1] == "(":
                        expr_stack.pop(-1)
                        break
                    else:
                        break
                self.op = args[::-1]

            else:

                self.op = self.parse(expr_stack)
                expr_stack.pop(-1)

        elif any([op == "True", op == "true", op == "False", op == "false"]):

            # return boolean
            self.op = True if op in "Truetrue" else False

        elif any(["float" in op, "bool" in op, "int" in op, "complex" in op]):

            expr_stack.pop(-1)

            # extract data type
            try:
                self.op = self.backend.add_op('cast', self.parse(expr_stack), op[0:-1], **self.parser_kwargs)
            except AttributeError:
                raise AttributeError(f"Datatype casting error in expression: {self.expr_str}. "
                                     f"{op[0:-1]} is not a valid data-type for this parser.")

        elif "." in op:

            self.op = float(op)

        elif op.isnumeric():

            self.op = int(op)

        elif op[0].isalpha():

            if self._finished_rhs:

                if op == 'rhs':
                    self.op = self.rhs
                else:
                    shape = self.rhs.shape if hasattr(self.rhs, 'shape') else ()
                    dtype = self.rhs.dtype if hasattr(self.rhs, 'dtype') else type(self.rhs)
                    self.op = self.backend.add_var(vtype='state_var', name=op, shape=shape, dtype=dtype,
                                                   **self.parser_kwargs)
                    self.vars[op] = self.op

            elif op == 't':

                self.op = self.backend.add_var(vtype='state_var', name=op, shape=(), dtype='float', value=0.0,
                                               **self.parser_kwargs)

            else:

                raise ValueError(f"Undefined variable detected in expression: {self.expr_str}. {op} was not found "
                                 f"in the respective arguments dictionary.")

        else:

            raise ValueError(f"Undefined operation detected in expression: {self.expr_str}. {op} cannot be "
                             f"interpreted by this parser.")

        return self.op

    def clear(self):
        """Clears expression list and stack.
        """
        self.expr_list.clear()
        self.expr_stack.clear()

    def _update_lhs(self):
        """Applies update to left-hand side of equation. For differential equations, different solving schemes are
        available.
        """

        # update left-hand side of equation
        ###################################

        diff_eq = self._diff_eq

        if diff_eq:

            lhs = self.vars[self.lhs_key]

            # update state variable vectors
            y_idx = self._append_to_var(var_name='y', val=lhs)
            self._append_to_var(var_name='y_delta', val=lhs)

            # extract left-hand side variable from state variable vector
            lhs_indexed = self.backend._create_op('index', self.backend.ops['index']['name'], self.vars['y'], y_idx)
            lhs_indexed.short_name = lhs.short_name

            if 'y_' in lhs_indexed.value:
                del_start, del_end = lhs_indexed.value.index('_'), lhs_indexed.value.index('[')
                lhs_indexed.value = lhs_indexed.value[:del_start] + lhs_indexed.value[del_end:]
            self.vars[self.lhs_key] = lhs_indexed

            # assign rhs to state var delta vector
            self.rhs = self.backend.add_op('=', self.vars['y_delta'], self.rhs, y_idx, **self.parser_kwargs)

            # broadcast rhs and lhs and store results in backend
            self.backend.state_vars.append(lhs.name)
            self.backend.vars[lhs.name] = self.vars[self.lhs_key]
            self.rhs.state_var = lhs.name

        else:

            # simple update
            if not self._instantaneous:
                self.backend.next_layer()
            indexed_lhs = "]" in self.expr_stack
            self.lhs = self.parse(self.expr_stack + ['rhs', self._assign_type])
            if not indexed_lhs:
                self.backend.lhs_vars.append(self.vars[self.lhs_key].name)
            if not self._instantaneous:
                self.backend.previous_layer()

    def _preprocess_expr_str(self, expr: str) -> tuple:
        """Turns differential equations into simple algebraic equations using a certain solver scheme and extracts
        left-hand side, right-hand side and update type of the equation.

        Parameters
        ----------
        expr
            Equation in string format.

        Returns
        -------
        tuple
            Contains left hand side, right hand side and left hand side update type
        """

        # collect equation specifics
        ############################

        # split equation into lhs and rhs and assign type
        lhs, rhs, assign_type = split_equation(expr)

        if not assign_type:
            return self._preprocess_expr_str(f"x = {expr}")

        # for the left-hand side, check whether it includes a differential operator
        if "d/dt" in lhs:
            diff_eq = True
            lhs_split = lhs.split('*')
            lhs = "".join(lhs_split[1:])
        elif "'" in lhs:
            diff_eq = True
            lhs = lhs.replace("'", "")
        elif "d" in lhs and "/dt" in lhs:
            diff_eq = True
            lhs = lhs.split('/dt')[0]
            lhs = lhs.replace("d", "", count=1)
        else:
            diff_eq = False

        # get clean name of lhs
        lhs_key = lhs.split('[')[0]
        lhs_key = lhs_key.replace(' ', '')
        lhs = lhs.replace(' ', '')

        # store equation specifics
        if diff_eq and assign_type != '=':
            raise ValueError(f'Wrong assignment method for equation: {expr}. '
                             f'A differential equation cannot be combined with an assign type other than `=`.')

        return lhs, rhs, diff_eq, assign_type, lhs_key

    def _push_first(self, strg, loc, toks):
        """Push tokens in first-to-last order to expression stack.
        """
        self.expr_stack.append(toks[0])

    def _push_neg(self, strg, loc, toks):
        """Push negative one multiplier if on first position in toks.
        """
        if toks and toks[0] == '-':
            self.expr_stack.append('-one')

    def _push_neg_or_first(self, strg, loc, toks):
        """Push neg one multipler to expression stack if on first position in toks, else push toks from first-to-last.
        """
        if toks and toks[0] == '-':
            self.expr_stack.append('-one')
        else:
            self.expr_stack.append(toks[0])

    def _push_last(self, strg, loc, toks):
        """Push tokens in last-to-first order to expression stack.
        """
        self.expr_stack.append(toks[-1])

    def _apply_idx(self, op: tp.Any, idx: tp.Any, update: tp.Optional[tp.Any] = None,
                   update_type: tp.Optional[str] = None, **kwargs) -> tp.Any:
        """Apply index idx to operation op.

        Parameters
        ----------
        op
            Operation to be indexed.
        idx
            Index to op.
        update
            Update to apply to op at idx.
        update_type
            Type of left-hand side update (e.g. `=` or `+=`).
        kwargs
            Additional keyword arguments to be passed to the indexing functions.

        Returns
        -------
        tp.Any
            Result of applying idx to op.

        """

        kwargs.update(self.parser_kwargs)

        # get constants/variables that are part of the index
        args = []
        if idx in self.vars['idx']:
            idx = self.vars['idx'].pop(idx)
        if type(idx) is str:
            idx_old = idx
            idx = []
            for idx_tmp in idx_old.split(','):
                for idx_tmp2 in idx_tmp.split(':'):
                    idx.append(idx_tmp2)
                    if idx_tmp2 in self.vars['idx']:
                        idx_var = self.vars['idx'].pop(idx_tmp2)
                        if not hasattr(idx_var, 'short_name'):
                            if hasattr(idx_var, 'shape') and tuple(idx_var.shape):
                                idx_var = idx_var[0]
                            idx[-1] = f"{idx_var}"
                        else:
                            if "_evaluated" in idx_var.short_name:
                                idx[-1] = f"{idx_var.numpy()}"
                            else:
                                idx[-1] = idx_var.short_name
                                args.append(idx_var)
                    idx.append(':')
                idx.pop(-1)
                idx.append(',')
            idx.pop(-1)
            idx = "".join(idx)

        return self.backend.apply_idx(op, idx, update, update_type, *tuple(args))

    def _check_parsed_expr(self, expr_str: str) -> None:
        """check whether parsing of expression string was successful.

        Parameters
        ----------
        expr_str
            Expression that has been attempted to be parsed.
        """

        for sub_str in sorted(self.expr_stack, key=len)[::-1]:
            if sub_str == 'E':
                sub_str = 'e'
            expr_str = expr_str.replace(sub_str, "")
        expr_str = expr_str.replace(" ", "")
        expr_str = expr_str.replace("(", "")
        expr_str = expr_str.replace(")", "")
        expr_str = expr_str.replace("-", "")
        if len(expr_str) > 0:
            raise ValueError(f"Error while parsing expression: {self.expr_str}. {expr_str} could not be parsed.")

    def _append_to_var(self, var_name: str, val: tp.Any) -> str:

        # create variable vector if not existent
        if var_name not in self.vars:
            self.vars[var_name] = self.backend.add_var(vtype='state_var', name=var_name, shape=(), dtype=val.dtype,
                                                       value=[], squeeze=False)

        # append left-hand side variable to variable vector
        var = self.backend.remove_var(var_name)
        var_val = var.numpy().tolist()
        append_val = val.numpy().tolist()
        if sum(var.shape) and sum(val.shape):
            new_val = var_val + append_val
        elif sum(var.shape):
            new_val = var_val + [append_val]
        else:
            new_val = append_val if type(append_val) is list else [append_val]

        # add updated variable to backend
        self.vars[var_name] = self.backend.add_var(vtype='state_var', name=var_name, value=new_val,
                                                   shape=(len(new_val),), dtype=var.dtype, squeeze=False)

        # return index to variable vector to retrieve appended values
        i1 = len(var_val) + self.backend.idx_start
        i2 = len(new_val) + self.backend.idx_start
        return f"{i1}:{i2}" if i2 - i1 > 1 else f"{i1}"

    @staticmethod
    def _compare(x: tp.Any, y: tp.Any) -> bool:
        """Checks whether x and y are equal or not.
        """
        test = x == y
        if hasattr(test, 'shape'):
            test = test.any()
        return test


def parse_equations(equations: list, equation_args: dict, backend: tp.Any, **kwargs) -> dict:
    """Parses a system (list) of equations into the backend. Transforms differential equations into the appropriate set
    of right-hand side evaluations that can be solved later on.

    Parameters
    ----------
    equations
        Collection of equations that describe the dynamics of the nodes and edges.
    equation_args
        Key-value pairs of arguments needed for parsing the equations.
    backend
        Backend instance to parse the equations into.
    kwargs
        Additional keyword arguments to be passed to the backend methods.

    Returns
    -------
    dict
        The updated equations args (in-place manipulation of all variables in equation_args happens during
        equation parsing).

    """
    state_vars = {}
    var_map = {}

    for layer in equations:
        for eq, scope in layer:

            # parse arguments
            #################

            # extract operator variables from equation args
            op_args = {key.split('/')[-1]: var for key, var in equation_args.items() if scope in key}
            inputs = op_args['inputs'] if 'inputs' in op_args else {}
            for key, inp in inputs.items():
                if inp not in equation_args:
                    raise KeyError(inp)
                if inp in var_map:
                    inp_tmp = var_map[inp]
                else:
                    inp_tmp = state_vars[inp] if inp in state_vars else equation_args[inp]
                    if type(inp_tmp) is dict:
                        inp_tmp = parse_dict({key: inp_tmp}, backend, scope="/".join(inp.split('/')[:-1]),
                                             **kwargs)[key]
                op_args[key] = inp_tmp

            # parse operator variables in backend
            args_tmp = {}
            for key, arg in op_args.items():
                if f"{scope}/{key}" in var_map:
                    op_args[key] = var_map[f"{scope}/{key}"]
                elif type(arg) is dict and 'vtype' in arg:
                    args_tmp[key] = arg
            args_tmp = parse_dict(args_tmp, backend, scope=scope, **kwargs)
            op_args.update(args_tmp)

            # add state variable vector to op args
            if 'y' in equation_args:
                op_args['y'] = equation_args['y']
                op_args['y_delta'] = equation_args['y_delta']

            # remember state variables
            for key, var in op_args.items():
                var_name = f"{scope}/{key}"
                if var_name not in var_map:
                    var_map[var_name] = var

            # parse equation
            ################

            instantaneous = is_diff_eq(eq) is False
            parser = ExpressionParser(expr_str=eq, args=op_args, backend=backend, scope=scope,
                                      instantaneous=instantaneous, **kwargs.copy())
            _, _, variables = parser.parse_expr()

            # update equations args
            #######################

            # save backend variables to equation args
            for key, var in variables.items():
                var_name = key if key == 'y' or key == 'y_delta' else f"{scope}/{key}"
                _, state_var = backend._is_state_var(var_name)
                if state_var and var_name not in state_vars:
                    state_vars[var_name] = var
                elif 'inputs' in variables and key not in variables['inputs']:
                    equation_args[var_name] = var

        # go to next layer in backend
        backend.add_layer()

    # save state variables in backend
    equation_args.update(state_vars)
    backend.vars.update(state_vars)

    return equation_args


def update_rhs(equations: list, equation_args: dict, update_num: int, update_str: str) -> tuple:
    """Update the right-hand side of all equations according to `update_str` and `update_num`. All state-variable
    occurrences will be replaced with the expression in the `update_str` template. Convenience function for differential
    equation solver that involve multiple partial updates of the state variables.

    Parameters
    ----------
    equations
        List of equations to be updated.
    equation_args
        Key-argument pairs of all relevant variables which occur in the equations.
    update_num
        Number of the partial update step for which the equations should be updated.
    update_str
        Template for the state variable replacement procedure. Should contain the following character strings:
        - `var_placeholder` will be replaced with the name of the state variables.
        - `var_placeholder_i` for partial updates of the state variables with `i` being a counter that needs to be
           replaced with the appropriate number of the partial update. Should be included for each partial update from
           i=1 to i=`update_num`.
        - `update_placeholder` for the position of the new, updated variable.

    Returns
    -------
    tuple
        List of the updated equations and dictionary with the equation arguments.

    """

    # collect variable updates from earlier rhs evaluations
    #######################################################

    var_updates = {}
    if update_num > 1:

        for key, arg in equation_args.items():

            # extraction of variable name
            node, op, var = key.split('/')
            if f"_upd_{update_num}" in var:
                var = var.replace(f"_upd_{update_num}", "")

            if "_upd_" in var:

                # find variable update number and cut of variable update identifier from variable name
                idx = int(var[-1])
                var_tmp = var[:-6]

                # indicate which variable placeholder to replace with this variable update below
                if var_tmp in var_updates:
                    var_updates[var_tmp].append((f'var_placeholder_{idx}', var))
                else:
                    var_updates[var_tmp] = [(f'var_placeholder_{idx}', var)]

    # integrate previous rhs evaluations into rhs equations
    #######################################################

    updated_args = {}
    while equation_args:

        key, arg = equation_args.popitem()

        if 'inputs' in key:

            # process input variables
            for var, arg_tmp in arg.copy().items():

                # extract variable name
                if f"_upd_{update_num}" in var:
                    var = var.replace(f"_upd_{update_num}", "")

                if "_upd_" not in var:

                    # create new variable name with update identifier
                    new_var = f"{var}_upd_{update_num}"
                    arg_tmp = arg_tmp.split('/')
                    arg_tmp[-1] = f"{arg_tmp[-1]}_upd_{update_num}"
                    arg_tmp = "/".join(arg_tmp)

                    if arg_tmp in equation_args or arg_tmp in updated_args:

                        # add variable update to inputs field
                        arg[new_var] = arg_tmp

                        # individualize the replacement template with variable name infos
                        replace_str = update_str.replace('update_placeholder', new_var)
                        if var in var_updates:
                            for placeholder, var_tmp in var_updates[var]:
                                replace_str = replace_str.replace(placeholder, var_tmp)
                        else:
                            for i in range(1, update_num):
                                replace_str = replace_str.replace(f'var_placeholder_{i}', f'{var}_upd_{i}')
                        replace_str = replace_str.replace('var_placeholder', var)

                        # go through equations and replace variable names with the individualized replacement template
                        for i, layer in enumerate(equations.copy()):
                            for j, (eq, scope) in enumerate(layer):
                                lhs, rhs, assign = split_equation(eq)
                                if replace_str not in rhs:
                                    rhs = replace(rhs, var, replace_str)
                                equations[i][j] = (f"{lhs} {assign} {rhs}", scope)

        else:

            # extract variable name
            node, op, var = key.split('/')
            if f"_upd_{update_num}" in var:
                var = var.replace(f"_upd_{update_num}", "")

            if "_upd_" in var:

                # save the variable to the arguments dictionary
                updated_args[f"{node}/{op}/{var}"] = arg

            else:

                # create new variable name and save the variable to the arguments dictionary
                new_var = f"{var}_upd_{update_num}"
                updated_args[f"{node}/{op}/{new_var}"] = arg

                # individualize the replacement template
                replace_str = update_str.replace('update_placeholder', new_var)
                if var in var_updates:
                    for placeholder, var_tmp in var_updates[var]:
                        replace_str = replace_str.replace(placeholder, var_tmp)
                replace_str = replace_str.replace('var_placeholder', var)

                # go through equations and replace variable occurences with the individualized replacement template
                for i, layer in enumerate(equations.copy()):
                    for j, (eq, scope) in enumerate(layer):
                        lhs, rhs, assign = split_equation(eq)
                        if replace_str not in rhs:
                            rhs = replace(rhs, var, replace_str)
                        equations[i][j] = (f"{lhs} {assign} {rhs}", scope)

    return equations, updated_args


def update_lhs(equations: list, equation_args: dict, update_num: int, var_dict: dict) -> tuple:
    """Update the left-hand side of all equations according to `update_num`. An update identifier will be added to all
    left-hand side state-variable occurences. Convenience function for differential equation solver that involve
    multiple partial updates of the state variables.

    Parameters
    ----------
    equations
        Equations, whose left-hand sides should be updated.
    equation_args
        Key-argument pairs including all relevant left-hand side variables.
    update_num
        Number of the partial udpate of the left-hand side variables.
    var_dict
        Key-argument pairs including the configurations of all state variables (like shape, dtype, vtype and value).


    Returns
    -------
    tuple
        List of the updated equations and dictionary with the equation arguments.
    """

    updated_args = {}
    while equation_args:

        # extract variable
        key, arg = equation_args.popitem()
        node, op, var = key.split('/')

        if "_upd_" in var and f"_upd_{update_num-1}" not in var:

            # add previous updates to the arguments dictionary
            updated_args[key] = arg

        else:

            # create new variable name
            var = var.replace(f"_upd_{update_num-1}", "")
            new_var = f"{var}_upd_{update_num}"

            # go through equations and replace the left-hand side variables of differential equations
            add_to_args = False
            for i, layer in enumerate(equations.copy()):
                for j, (eq, scope) in enumerate(layer):
                    if node in scope and op in scope:
                        lhs, rhs, _ = split_equation(eq)
                        if var in lhs and new_var in replace(lhs, var, new_var):
                            de = False
                            if "d/dt" in lhs:
                                de = True
                                add_to_args = True
                                lhs = lhs.replace("d/dt", "")
                                lhs = lhs.replace("*", "")
                                lhs = lhs.replace(" ", "")
                            elif "'" in lhs:
                                de = True
                                add_to_args = True
                                lhs = lhs.replace("'", "")
                                lhs = lhs.replace(" ", "")
                            if de:
                                lhs = replace(lhs, var, new_var)
                                equations[i][j] = (f"{lhs} = step_size * ({rhs})", scope)

            if add_to_args:

                # save updated left-hand side variable to arguments dictionary
                for var_key, var in var_dict.copy().items():
                    if var_key == key or f"{var_key}_upd_" in key:
                        arg = var
                        break
                updated_args[f"{node}/{op}/{new_var}"] = arg

    return equations, updated_args


def update_equation_args(args: dict, updates: dict) -> dict:
    """Save variable updates to the equation args dictionary.

    Parameters
    ----------
    args
        Equation argument dictionary.
    updates
        Dictionary with variable updates.

    Returns
    -------
    dict
        Updated equation argument dictionary.
    """

    args_new = {}

    # add variables updates to equation arguments
    for key, arg in args.items():
        if key in updates:
            args_new[key] = updates[key]
        else:
            args_new[key] = arg

    # check which input fields need to be updated as well
    inputs = [key for key in args if 'inputs' in key]
    for inp in inputs:
        for in_key, in_map in args[inp].copy().items():
            for upd in updates:
                if in_map in upd:
                    args_new[inp].update({upd.split('/')[-1]: upd})
                    break

    return args_new


def parse_dict(var_dict: dict, backend, **kwargs) -> dict:
    """Parses a dictionary with variable information and creates backend variables from that information.

    Parameters
    ----------
    var_dict
        Contains key-value pairs for each variable that should be translated into the backend graph.
        Each value is a dictionary again containing the variable information (needs at least a field for `vtype`).
    backend
        Backend instance that the variables should be added to.
    kwargs
        Additional keyword arguments to be passed to the backend methods.

    Returns
    -------
    dict
        Key-value pairs with the backend variable names and handles.
    """

    var_dict_new = {}

    # go through dictionary items and instantiate variables
    #######################################################

    for var_name, var in var_dict.items():

        # instantiate variable
        if var['vtype'] == 'raw':
            var_dict_new[var_name] = var['value']
        else:
            var.update(kwargs)
            if var_name == 'y' or var_name == 'y_delta':
                raise KeyError(f'Warning: Variable name {var_name} is reserved for pyrates-internal state variables. '
                               f'Please choose a different variable name.')
            var_dict_new[var_name] = backend.add_var(name=var_name, **var)

    return var_dict_new


def split_equation(expr: str) -> tuple:
    """Splits an equation string into a left-hand side, right-and side and an assign type.

    Parameters
    ----------
    expr
        Equation string. Should contain a left-hand side and a right-hand side, separated by some form of assign symbol.

    Returns
    -------
    tuple
        left-hand side string, right-hand side string, assign operation string.
    """

    # define assign types and explicit non-assign types
    assign_types = ['+=', '-=', '*=', '/=']
    not_assign_types = ['<=', '>=', '==', '!=']

    lhs, rhs, assign_type, found_assign_type = "", "", "", False

    # look for assign types in expression
    for assign_type in assign_types:

        if assign_type in expr:

            # split expression via assign symbol
            if f' {assign_type} ' in expr:
                lhs, rhs = expr.split(f' {assign_type} ', maxsplit=1)
            elif f' {assign_type}' in expr:
                lhs, rhs = expr.split(f' {assign_type}', maxsplit=1)
            elif f'{assign_type} ' in expr:
                lhs, rhs = expr.split(f'{assign_type} ', maxsplit=1)
            else:
                lhs, rhs = expr.split(assign_type, maxsplit=1)
            found_assign_type = True
            break

        elif '=' in expr:

            # assume standard assign
            assign_type = '='
            assign = True

            # check if `=` symbol marks an assign operation or not
            for not_assign_type in not_assign_types:
                if not_assign_type in expr:
                    expr_tmp = expr.replace(not_assign_type, '')
                    if '=' not in expr_tmp:
                        assign = False

            if assign:

                # split equation via `=` symbol
                if f' = ' in expr:
                    lhs, rhs = expr.split(f' = ', maxsplit=1)
                elif f' {assign_type}' in expr:
                    lhs, rhs = expr.split(f' =', maxsplit=1)
                elif f'{assign_type} ' in expr:
                    lhs, rhs = expr.split(f'= ', maxsplit=1)
                else:
                    lhs, rhs = expr.split(f"=", maxsplit=1)
                found_assign_type = True
                break

    if not found_assign_type:
        return lhs, rhs, False
    return lhs, rhs, assign_type


def replace(eq: str, term: str, replacement: str, rhs_only: tp.Optional[bool] = False,
            lhs_only: tp.Optional[bool] = False) -> str:
    """Replaces a term in an equation with a replacement term (save replacement).

    Parameters
    ----------
    eq
        Equation that includes the term.
    term
        Term that should be replaced.
    replacement
        Replacement for all occurences of term.
    rhs_only
        If True, replacements will only be performed in right-hand side of the equation.
    lhs_only
        IF True, replacements will only be performed in left-hand side of the equation.

    Returns
    -------
    str
        The updated equation.
    """

    # define follow-up operations/signs that are allowed to follow directly after term in eq
    allowed_follow_ops = '+=*/^<>=!.%@[]():, '

    # replace every proper appearance of term in eq with replacement
    ################################################################

    eq_new = ""
    idx = eq.find(term)

    # go through all appearances of term in eq
    while idx != -1:

        # get idx of sign that follows after term
        idx_follow_op = idx+len(term)

        # if it is an allowed sign, replace term, else not
        replaced = False
        if ((idx_follow_op < len(eq) and eq[idx_follow_op] in allowed_follow_ops) and
           (idx == 0 or eq[idx-1] in allowed_follow_ops)) or \
                (idx_follow_op == len(eq) and eq[idx-1] in allowed_follow_ops):
            eq_part = eq[:idx]
            if (rhs_only and "=" in eq_part) or (lhs_only and "=" not in eq_part) or (not rhs_only and not lhs_only):
                eq_new += f"{eq_part} {replacement}"
                replaced = True
        if not replaced:
            eq_new += f"{eq[:idx_follow_op]}"

        # jump to next appearance of term in eq
        eq = eq[idx_follow_op:]
        idx = eq.find(term)

    # add rest of eq to new eq
    eq_new += eq

    return eq_new


def is_diff_eq(eq: str) -> bool:
    """Checks whether `eq` is a differential equation or not.

    Parameters
    ----------
    eq
        Equation string.

    Returns
    -------
    bool
        True, if `eq` is a differential equation.

    """

    lhs, rhs, _ = split_equation(eq)
    if "d/dt" in lhs:
        de = True
    elif "'" in lhs:
        de = True
    elif "dt_test" in replace(rhs, "step_size", "dt_test"):
        de = True
    else:
        de = False

    return de


def is_coupled(eqs: list) -> bool:
    """Checks whether a list of equations defines a set of coupled equations, i.e. at least one left-hand side variable
    appears in the right-hand side of another equation.

    Parameters
    ----------
    eqs
        List of equation strings

    Returns
    -------
    bool
        True, if at least one left-hand side variable appears in the right-hand side of another equation.
    """

    # extract lhs, rhs and scope of each equation
    lhs_col, rhs_col, scope_col = [], [], []
    for eq, scope in eqs:
        lhs, rhs, _ = split_equation(eq)
        lhs_col.append(lhs)
        rhs_col.append(rhs)
        scope_col.append(scope)

    # check whether equations are coupled
    for lhs, lhs_scope in zip(lhs_col, scope_col):
        lhs = lhs.replace(" ", "")
        lhs = lhs.replace("d/dt", "")
        lhs = lhs.replace("*", "")
        lhs = lhs.replace("'", "")
        if "[" in lhs:
            l_idx = lhs.index("[")
            r_idx = lhs.index("]")
            lhs = lhs.replace(lhs[l_idx:r_idx+1], "")
        for rhs, rhs_scope in zip(rhs_col, scope_col):
            if "__test_string__" in replace(rhs, lhs, "__test_string__") and rhs_scope == lhs_scope:
                return True

    return False
